"""
Test the jinja module
"""

import os

import salt.utils.files
import salt.utils.json
import salt.utils.yaml
from tests.support.runtests import RUNTIME_VARS


def _path(name, absolute=False):
    path = os.path.join("modules", "jinja", name)
    if absolute:
        return os.path.join(RUNTIME_VARS.BASE_FILES, path)
    else:
        return path


def test_import_json(salt_cli, salt_minion):
    json_file = "osarchmap.json"
    ret = salt_cli.run("jinja.import_json", _path(json_file), minion_tgt=salt_minion.id)
    with salt.utils.files.fopen(_path(json_file, absolute=True)) as fh_:
        assert salt.utils.json.load(fh_) == ret.data


def test_import_yaml(salt_cli, salt_minion):
    yaml_file = "defaults.yaml"
    ret = salt_cli.run("jinja.import_yaml", _path(yaml_file), minion_tgt=salt_minion.id)
    with salt.utils.files.fopen(_path(yaml_file, absolute=True)) as fh_:
        assert salt.utils.yaml.safe_load(fh_) == ret.data


def test_load_map(grains, salt_cli, salt_minion):
    ret = salt_cli.run(
        "jinja.load_map", _path("map.jinja"), "template", minion_tgt=salt_minion.id
    )

    assert isinstance(
        ret.data, dict
    ), f"failed to return dictionary from jinja.load_map: {ret}"

    with salt.utils.files.fopen(_path("defaults.yaml", absolute=True)) as fh_:
        defaults = salt.utils.yaml.safe_load(fh_)
    with salt.utils.files.fopen(_path("osarchmap.json", absolute=True)) as fh_:
        osarchmap = salt.utils.json.load(fh_)
    with salt.utils.files.fopen(_path("osfamilymap.yaml", absolute=True)) as fh_:
        osfamilymap = salt.utils.yaml.safe_load(fh_)
    with salt.utils.files.fopen(_path("osmap.yaml", absolute=True)) as fh_:
        osmap = salt.utils.yaml.safe_load(fh_)
    with salt.utils.files.fopen(_path("osfingermap.yaml", absolute=True)) as fh_:
        osfingermap = salt.utils.yaml.safe_load(fh_)

    assert ret.data.get("arch") == osarchmap.get(grains["osarch"], {}).get("arch")
    assert ret.data.get("config") == osfingermap.get(grains["osfinger"], {}).get(
        "config",
        osmap.get(grains["os"], {}).get(
            "config",
            osfamilymap.get(grains["os_family"], {}).get(
                "config", defaults.get("template").get("config")
            ),
        ),
    )



import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module")
def pkg_def_contents(state_tree):
    return r"""
    my-software:
      '1.0.1':
        full_name: 'My Software'
        installer: 'C:\files\mysoftware.msi'
        install_flags: '/qn /norestart'
        uninstaller: 'C:\files\mysoftware.msi'
        uninstall_flags: '/qn /norestart'
        msiexec: True
        reboot: False
    """


@pytest.fixture(scope="module")
def pkg(modules):
    yield modules.pkg


def test_refresh_db(pkg, pkg_def_contents, state_tree, minion_opts):
    assert len(pkg.get_package_info("my-software")) == 0
    repo_dir = state_tree / "winrepo_ng"
    with pytest.helpers.temp_file("my-software.sls", pkg_def_contents, repo_dir):
        pkg.refresh_db()
    assert len(pkg.get_package_info("my-software")) == 1



"""
Support for VictorOps

.. versionadded:: 2015.8.0

Requires an ``api_key`` in ``/etc/salt/minion``:

.. code-block:: yaml

    victorops:
      api_key: '280d4699-a817-4719-ba6f-ca56e573e44f'
"""

import datetime
import logging
import time

import salt.utils.http
import salt.utils.json
from salt.exceptions import SaltInvocationError

log = logging.getLogger(__name__)

_api_key_missing_error = "No VictorOps api key found."


def __virtual__():
    """
    Only load the module if apache is installed
    """
    if not __salt__["config.get"]("victorops.api_key") and not __salt__["config.get"](
        "victorops:api_key"
    ):
        return (False, _api_key_missing_error)
    return True


def _query(
    action=None, routing_key=None, args=None, method="GET", header_dict=None, data=None
):
    """
    Make a web call to VictorOps
    """
    api_key = __salt__["config.get"]("victorops.api_key") or __salt__["config.get"](
        "victorops:api_key"
    )

    path = "https://alert.victorops.com/integrations/generic/20131114/"

    if action:
        path += f"{action}/"

    if api_key:
        path += f"{api_key}/"

    if routing_key:
        path += routing_key

    log.debug("VictorOps URL: %s", path)

    if not isinstance(args, dict):
        args = {}

    if header_dict is None:
        header_dict = {"Content-type": "application/json"}

    if method != "POST":
        header_dict["Accept"] = "application/json"

    decode = True
    if method == "DELETE":
        decode = False

    result = salt.utils.http.query(
        path,
        method,
        params=args,
        data=data,
        header_dict=header_dict,
        decode=decode,
        decode_type="json",
        text=True,
        status=True,
        cookies=True,
        persist_session=True,
        opts=__opts__,
    )
    if "error" in result:
        log.error(result["error"])
        return [result["status"], result["error"]]

    return [result["status"], result.get("dict", {})]


def create_event(message_type=None, routing_key="everybody", **kwargs):
    """
    Create an event in VictorOps. Designed for use in states.

    The following parameters are required:

    :param message_type:            One of the following values: INFO, WARNING, ACKNOWLEDGEMENT, CRITICAL, RECOVERY.

    The following parameters are optional:

    :param routing_key:             The key for where messages should be routed. By default, sent to
                                    'everyone' route.

    :param entity_id:               The name of alerting entity. If not provided, a random name will be assigned.

    :param timestamp:               Timestamp of the alert in seconds since epoch. Defaults to the
                                    time the alert is received at VictorOps.

    :param timestamp_fmt            The date format for the timestamp parameter.

    :param state_start_time:        The time this entity entered its current state
                                    (seconds since epoch). Defaults to the time alert is received.

    :param state_start_time_fmt:    The date format for the timestamp parameter.

    :param state_message:           Any additional status information from the alert item.

    :param entity_is_host:          Used within VictorOps to select the appropriate
                                    display format for the incident.

    :param entity_display_name:     Used within VictorOps to display a human-readable name for the entity.

    :param ack_message:             A user entered comment for the acknowledgment.

    :param ack_author:              The user that acknowledged the incident.

    :return:                        A dictionary with result, entity_id, and message if result was failure.

    CLI Example:

    .. code-block:: yaml

        salt myminion victorops.create_event message_type='CRITICAL' routing_key='everyone' \
                 entity_id='hostname/diskspace'

        salt myminion victorops.create_event message_type='ACKNOWLEDGEMENT' routing_key='everyone' \
                 entity_id='hostname/diskspace' ack_message='Acknowledged' ack_author='username'

        salt myminion victorops.create_event message_type='RECOVERY' routing_key='everyone' \
                 entity_id='hostname/diskspace'

    The following parameters are required:
        message_type

    """

    keyword_args = {
        "entity_id": str,
        "state_message": str,
        "entity_is_host": bool,
        "entity_display_name": str,
        "ack_message": str,
        "ack_author": str,
    }

    data = {}

    if not message_type:
        raise SaltInvocationError('Required argument "message_type" is missing.')

    if message_type.upper() not in [
        "INFO",
        "WARNING",
        "ACKNOWLEDGEMENT",
        "CRITICAL",
        "RECOVERY",
    ]:
        raise SaltInvocationError(
            '"message_type" must be INFO, WARNING, ACKNOWLEDGEMENT, CRITICAL, or'
            " RECOVERY."
        )

    data["message_type"] = message_type

    data["monitoring_tool"] = "SaltStack"

    if "timestamp" in kwargs:
        timestamp_fmt = kwargs.get("timestamp_fmt", "%Y-%m-%dT%H:%M:%S")

        try:
            timestamp = datetime.datetime.strptime(kwargs["timestamp"], timestamp_fmt)
            data["timestamp"] = int(time.mktime(timestamp.timetuple()))
        except (TypeError, ValueError):
            raise SaltInvocationError(
                "Date string could not be parsed: {}, {}".format(
                    kwargs["timestamp"], timestamp_fmt
                )
            )

    if "state_start_time" in kwargs:
        state_start_time_fmt = kwargs.get("state_start_time_fmt", "%Y-%m-%dT%H:%M:%S")

        try:
            state_start_time = datetime.datetime.strptime(
                kwargs["state_start_time"], state_start_time_fmt
            )
            data["state_start_time"] = int(time.mktime(state_start_time.timetuple()))
        except (TypeError, ValueError):
            raise SaltInvocationError(
                "Date string could not be parsed: {}, {}".format(
                    kwargs["state_start_time"], state_start_time_fmt
                )
            )

    for kwarg in keyword_args:
        if kwarg in kwargs:
            if isinstance(kwargs[kwarg], keyword_args[kwarg]):
                data[kwarg] = kwargs[kwarg]
            else:
                # Should this faile on the wrong type.
                log.error("Wrong type, skipping %s", kwarg)

    status, result = _query(
        action="alert",
        routing_key=routing_key,
        data=salt.utils.json.dumps(data),
        method="POST",
    )
    return result



#
# Author: Bo Maryniuk <bo@suse.de>
#
# Copyright 2018 SUSE LLC
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import os
import tempfile

import pytest

import salt.utils.files
import salt.utils.stringutils
from salt.modules import x509
from tests.support.helpers import dedent
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

try:
    import M2Crypto  # pylint: disable=unused-import

    HAS_M2CRYPTO = True
except ImportError:
    HAS_M2CRYPTO = False


default_values = {
    "ca_key": b"""-----BEGIN RSA PRIVATE KEY-----
MIICWwIBAAKBgQCjdjbgL4kQ8Lu73xeRRM1q3C3K3ptfCLpyfw38LRnymxaoJ6ls
pNSx2dU1uJ89YKFlYLo1QcEk4rJ2fdIjarV0kuNCY3rC8jYUp9BpAU5Z6p9HKeT1
2rTPH81JyjbQDR5PyfCyzYOQtpwpB4zIUUK/Go7tTm409xGKbbUFugJNgQIDAQAB
AoGAF24we34U1ZrMLifSRv5nu3OIFNZHyx2DLDpOFOGaII5edwgIXwxZeIzS5Ppr
yO568/8jcdLVDqZ4EkgCwRTgoXRq3a1GLHGFmBdDNvWjSTTMLoozuM0t2zjRmIsH
hUd7tnai9Lf1Bp5HlBEhBU2gZWk+SXqLvxXe74/+BDAj7gECQQDRw1OPsrgTvs3R
3MNwX6W8+iBYMTGjn6f/6rvEzUs/k6rwJluV7n8ISNUIAxoPy5g5vEYK6Ln/Ttc7
u0K1KNlRAkEAx34qcxjuswavL3biNGE+8LpDJnJx1jaNWoH+ObuzYCCVMusdT2gy
kKuq9ytTDgXd2qwZpIDNmscvReFy10glMQJAXebMz3U4Bk7SIHJtYy7OKQzn0dMj
35WnRV81c2Jbnzhhu2PQeAvt/i1sgEuzLQL9QEtSJ6wLJ4mJvImV0TdaIQJAAYyk
TcKK0A8kOy0kMp3yvDHmJZ1L7wr7bBGIZPBlQ0Ddh8i1sJExm1gJ+uN2QKyg/XrK
tDFf52zWnCdVGgDwcQJALW/WcbSEK+JVV6KDJYpwCzWpKIKpBI0F6fdCr1G7Xcwj
c9bcgp7D7xD+TxWWNj4CSXEccJgGr91StV+gFg4ARQ==
-----END RSA PRIVATE KEY-----
""",
    "x509_args_ca": {
        "text": True,
        "CN": "Redacted Root CA",
        "O": "Redacted",
        "C": "BE",
        "ST": "Antwerp",
        "L": "Local Town",
        "Email": "certadm@example.org",
        "basicConstraints": "critical CA:true",
        "keyUsage": "critical cRLSign, keyCertSign",
        "subjectKeyIdentifier": "hash",
        "authorityKeyIdentifier": "keyid,issuer:always",
        "days_valid": 3650,
        "days_remaining": 0,
    },
    "x509_args_cert": {
        "text": True,
        "CN": "Redacted Normal Certificate",
        "O": "Redacted",
        "C": "BE",
        "ST": "Antwerp",
        "L": "Local Town",
        "Email": "certadm@example.org",
        "basicConstraints": "critical CA:false",
        "keyUsage": "critical keyEncipherment",
        "subjectKeyIdentifier": "hash",
        "authorityKeyIdentifier": "keyid,issuer:always",
    },
    "crl_args": {
        "text": False,
        "signing_private_key_passphrase": None,
        "revoked": None,
        "include_expired": False,
        "days_valid": 100,
        "digest": "sha512",
    },
}


class X509TestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {x509: {"__opts__": {"fips_mode": False}}}

    @patch("salt.modules.x509.log", MagicMock())
    def test_private_func__parse_subject(self):
        """
        Test private function _parse_subject(subject) it handles a missing fields
        :return:
        """

        class FakeSubject:
            """
            Class for faking x509'th subject.
            """

            def __init__(self):
                self.nid = {"Darth Vader": 1}

            def __getattr__(self, item):
                if item != "nid":
                    raise TypeError(
                        "A star wars satellite accidentally blew up the WAN."
                    )

        subj = FakeSubject()
        x509._parse_subject(subj)
        assert x509.log.trace.call_args[0][0] == "Missing attribute '%s'. Error: %s"
        assert x509.log.trace.call_args[0][1] == list(subj.nid.keys())[0]
        assert isinstance(x509.log.trace.call_args[0][2], TypeError)

    @pytest.mark.skipif(
        not HAS_M2CRYPTO, reason="Skipping, reason=M2Crypto is unavailable"
    )
    def test_get_pem_entry(self):
        """
        Test private function _parse_subject(subject) it handles a missing fields
        :return:
        """
        ca_key = default_values["ca_key"]
        ret = x509.get_pem_entry(ca_key)
        self.assertEqual(ret, ca_key)

    @pytest.mark.skipif(
        not HAS_M2CRYPTO, reason="Skipping, reason=M2Crypto is unavailable"
    )
    def test_get_private_key_size(self):
        """
        Test private function _parse_subject(subject) it handles a missing fields
        :return:
        """
        ca_key = default_values["ca_key"]
        ret = x509.get_private_key_size(ca_key)
        self.assertEqual(ret, 1024)

    @pytest.mark.skipif(
        not HAS_M2CRYPTO, reason="Skipping, reason=M2Crypto is unavailable"
    )
    def test_create_key(self):
        """
        Test that x509.create_key returns a private key
        :return:
        """
        ret = x509.create_private_key(text=True, passphrase="super_secret_passphrase")
        self.assertIn("BEGIN RSA PRIVATE KEY", ret)

    @pytest.mark.skipif(
        not HAS_M2CRYPTO, reason="Skipping, reason=M2Crypto is unavailable"
    )
    def test_create_certificate(self):
        """
        Test private function _parse_subject(subject) it handles a missing fields
        :return:
        """
        ca_key = default_values["ca_key"]
        ca_kwargs = default_values["x509_args_ca"].copy()
        ca_kwargs["signing_private_key"] = ca_key
        ret = x509.create_certificate(**ca_kwargs)
        self.assertIn("BEGIN CERTIFICATE", ret)

    @pytest.mark.skipif(
        not HAS_M2CRYPTO, reason="Skipping, reason=M2Crypto is unavailable"
    )
    def test_create_certificate_with_not_after(self):
        ca_key = default_values["ca_key"]
        ca_kwargs = default_values["x509_args_ca"].copy()
        ca_kwargs["signing_private_key"] = ca_key

        # Issue the CA certificate (self-signed)
        ca_cert = x509.create_certificate(**ca_kwargs)

        fmt = "%Y-%m-%d %H:%M:%S"
        # We also gonna use the current date in UTC format for verification
        not_after = datetime.datetime.utcnow()
        # And set the UTC timezone to the naive datetime resulting from parsing
        not_after = not_after.replace(tzinfo=M2Crypto.ASN1.UTC)
        not_after_str = datetime.datetime.strftime(not_after, fmt)

        # Sign a new server certificate with the CA
        ca_key = default_values["ca_key"]
        cert_kwargs = default_values["x509_args_cert"].copy()
        cert_kwargs["signing_private_key"] = ca_key
        cert_kwargs["signing_cert"] = ca_cert
        cert_kwargs["not_after"] = not_after_str
        server_cert = x509.create_certificate(**cert_kwargs)

        not_after_from_cert = ""
        # Save server certificate to disk so we can check its properties
        with tempfile.NamedTemporaryFile("w+") as server_cert_file:
            server_cert_file.write(salt.utils.stringutils.to_str(server_cert))
            server_cert_file.flush()

            # Retrieve not_after property from server certificate
            server_cert_details = x509.read_certificate(server_cert_file.name)
            not_after_from_cert = server_cert_details["Not After"]

        # Check if property is the one we've added to the certificate. The
        # property from the certificate will come as a string with no timezone
        # information in it.
        self.assertIn(not_after_str, not_after_from_cert)

    @pytest.mark.skipif(
        not HAS_M2CRYPTO, reason="Skipping, reason=M2Crypto is unavailable"
    )
    def test_create_certificate_with_not_before(self):
        ca_key = default_values["ca_key"]
        ca_kwargs = default_values.get("x509_args_ca").copy()
        ca_kwargs["signing_private_key"] = ca_key

        # Issue the CA certificate (self-signed)
        ca_cert = x509.create_certificate(**ca_kwargs)

        fmt = "%Y-%m-%d %H:%M:%S"
        # We also gonna use the current date in UTC format for verification
        not_before = datetime.datetime.utcnow()
        # And set the UTC timezone to the naive datetime resulting from parsing
        not_before = not_before.replace(tzinfo=M2Crypto.ASN1.UTC)
        not_before_str = datetime.datetime.strftime(not_before, fmt)

        # Sign a new server certificate with the CA
        ca_key = default_values["ca_key"]
        cert_kwargs = default_values["x509_args_cert"].copy()
        cert_kwargs["signing_private_key"] = ca_key
        cert_kwargs["signing_cert"] = ca_cert
        cert_kwargs["not_before"] = not_before_str
        server_cert = x509.create_certificate(**cert_kwargs)

        not_before_from_cert = ""
        # Save server certificate to disk so we can check its properties
        with tempfile.NamedTemporaryFile("w+") as server_cert_file:
            server_cert_file.write(salt.utils.stringutils.to_str(server_cert))
            server_cert_file.flush()
            # Retrieve not_after property from server certificate
            server_cert_details = x509.read_certificate(server_cert_file.name)
            not_before_from_cert = server_cert_details["Not Before"]

        # Check if property is the one we've added to the certificate. The
        # property will come from the certificate as a string with no timezone
        # information in it.
        self.assertIn(not_before_str, not_before_from_cert)

    @pytest.mark.skipif(
        not HAS_M2CRYPTO, reason="Skipping, reason=M2Crypto is unavailable"
    )
    def test_create_certificate_with_not_before_wrong_date(self):
        ca_key = default_values["ca_key"]
        ca_kwargs = default_values.get("x509_args_ca").copy()
        ca_kwargs["signing_private_key"] = ca_key

        # Issue the CA certificate (self-signed)
        ca_cert = x509.create_certificate(**ca_kwargs)

        not_before_str = "this is an intentionally wrong format"

        # Try to sign a new server certificate with the wrong date
        msg = (
            "not_before: this is an intentionally wrong format is not in required"
            " format %Y-%m-%d %H:%M:%S"
        )
        with self.assertRaisesRegex(salt.exceptions.SaltInvocationError, msg):
            ca_key = default_values["ca_key"]
            cert_kwargs = default_values["x509_args_cert"].copy()
            cert_kwargs["signing_private_key"] = ca_key
            cert_kwargs["signing_cert"] = ca_cert
            cert_kwargs["not_before"] = not_before_str
            x509.create_certificate(**cert_kwargs)

    @pytest.mark.skipif(
        not HAS_M2CRYPTO, reason="Skipping, reason=M2Crypto is unavailable"
    )
    def test_create_certificate_with_not_after_wrong_date(self):
        ca_key = default_values["ca_key"]
        ca_kwargs = default_values.get("x509_args_ca").copy()
        ca_kwargs["signing_private_key"] = ca_key

        # Issue the CA certificate (self-signed)
        ca_cert = x509.create_certificate(**ca_kwargs)

        not_after_str = "this is an intentionally wrong format"

        # Try to sign a new server certificate with the wrong date
        msg = (
            "not_after: this is an intentionally wrong format is not in required format"
            " %Y-%m-%d %H:%M:%S"
        )
        with self.assertRaisesRegex(salt.exceptions.SaltInvocationError, msg):
            ca_key = default_values["ca_key"]
            cert_kwargs = default_values["x509_args_cert"].copy()
            cert_kwargs["signing_private_key"] = ca_key
            cert_kwargs["signing_cert"] = ca_cert
            cert_kwargs["not_after"] = not_after_str
            x509.create_certificate(**cert_kwargs)

    @pytest.mark.skipif(
        not HAS_M2CRYPTO, reason="Skipping, reason=M2Crypto is unavailable"
    )
    def test_create_certificate_with_not_before_and_not_after(self):
        ca_key = default_values["ca_key"]
        ca_kwargs = default_values.get("x509_args_ca").copy()
        ca_kwargs["signing_private_key"] = ca_key

        # Issue the CA certificate (self-signed)
        ca_cert = x509.create_certificate(**ca_kwargs)

        fmt = "%Y-%m-%d %H:%M:%S"
        # Here we gonna use the current date as the not_before date
        # First we again take the UTC for verification
        not_before = datetime.datetime.utcnow()
        # And set the UTC timezone to the naive datetime resulting from parsing
        not_before = not_before.replace(tzinfo=M2Crypto.ASN1.UTC)
        not_before_str = datetime.datetime.strftime(not_before, fmt)
        # And we use the same logic to generate a not_after 5 days in the
        # future
        not_after = not_before + datetime.timedelta(days=5)
        # And set the UTC timezone to the naive datetime resulting from parsing
        not_after = not_after.replace(tzinfo=M2Crypto.ASN1.UTC)
        not_after_str = datetime.datetime.strftime(not_after, fmt)

        # Sign a new server certificate with the CA
        ca_key = default_values["ca_key"]
        cert_kwargs = default_values["x509_args_cert"].copy()
        cert_kwargs["signing_private_key"] = ca_key
        cert_kwargs["signing_cert"] = ca_cert
        cert_kwargs["not_after"] = not_after_str
        cert_kwargs["not_before"] = not_before_str
        server_cert = x509.create_certificate(**cert_kwargs)

        not_after_from_cert = ""
        not_before_from_cert = ""
        # Save server certificate to disk so we can check its properties
        with tempfile.NamedTemporaryFile("w+") as server_cert_file:
            server_cert_file.write(salt.utils.stringutils.to_str(server_cert))
            server_cert_file.flush()

            # Retrieve not_after property from server certificate
            server_cert_details = x509.read_certificate(server_cert_file.name)
            not_before_from_cert = server_cert_details["Not Before"]
            not_after_from_cert = server_cert_details["Not After"]

        # Check if property values are the ones we've added to the certificate.
        # The values will come as strings containing no timezone information in
        # them.
        self.assertIn(not_before_str, not_before_from_cert)
        self.assertIn(not_after_str, not_after_from_cert)

    @pytest.mark.skipif(
        not HAS_M2CRYPTO, reason="Skipping, reason=M2Crypto is unavailable"
    )
    def test_create_crl(self):
        ca_key = default_values["ca_key"]
        ca_kwargs = default_values.get("x509_args_ca").copy()
        ca_kwargs["signing_private_key"] = ca_key

        ca_cert = x509.create_certificate(**ca_kwargs)

        with tempfile.NamedTemporaryFile("w+", delete=False) as ca_key_file:
            ca_key_file.write(salt.utils.stringutils.to_str(ca_key))
            ca_key_file.flush()

        with tempfile.NamedTemporaryFile("w+", delete=False) as ca_cert_file:
            ca_cert_file.write(salt.utils.stringutils.to_str(ca_cert))
            ca_cert_file.flush()

        with tempfile.NamedTemporaryFile("w+", delete=False) as ca_crl_file:
            crl_kwargs = default_values.get("crl_args").copy()
            crl_kwargs["path"] = ca_crl_file.name
            crl_kwargs["signing_private_key"] = ca_key_file.name
            crl_kwargs["signing_cert"] = ca_cert_file.name
            x509.create_crl(**crl_kwargs)

        with salt.utils.files.fopen(ca_crl_file.name, "r") as crl_file:
            crl = crl_file.read()

        os.remove(ca_key_file.name)
        os.remove(ca_cert_file.name)
        os.remove(ca_crl_file.name)

        # Ensure that a CRL was actually created
        self.assertIn("BEGIN X509 CRL", crl)

    @pytest.mark.skipif(
        not HAS_M2CRYPTO, reason="Skipping, reason=M2Crypto is unavailable"
    )
    def test_revoke_certificate_with_crl(self):
        ca_key = default_values["ca_key"]
        ca_kwargs = default_values.get("x509_args_ca").copy()
        ca_kwargs["signing_private_key"] = ca_key

        # Issue the CA certificate (self-signed)
        ca_cert = x509.create_certificate(**ca_kwargs)

        # Sign a new server certificate with the CA
        ca_key = default_values["ca_key"]
        cert_kwargs = default_values["x509_args_cert"].copy()
        cert_kwargs["signing_private_key"] = ca_key
        cert_kwargs["signing_cert"] = ca_cert
        server_cert = x509.create_certificate(**cert_kwargs)

        # Save CA cert + key and server cert to disk as PEM files
        with tempfile.NamedTemporaryFile("w+", delete=False) as ca_key_file:
            ca_key_file.write(salt.utils.stringutils.to_str(ca_key))
            ca_key_file.flush()

        with tempfile.NamedTemporaryFile("w+", delete=False) as ca_cert_file:
            ca_cert_file.write(salt.utils.stringutils.to_str(ca_cert))
            ca_cert_file.flush()

        with tempfile.NamedTemporaryFile("w+", delete=False) as server_cert_file:
            server_cert_file.write(salt.utils.stringutils.to_str(server_cert))
            server_cert_file.flush()

        # Revoke server CRL
        revoked = [
            {
                "certificate": server_cert_file.name,
                "revocation_date": "2015-03-01 00:00:00",
            }
        ]
        with tempfile.NamedTemporaryFile("w+", delete=False) as ca_crl_file:
            crl_kwargs = default_values.get("crl_args").copy()
            crl_kwargs["path"] = ca_crl_file.name
            crl_kwargs["signing_private_key"] = ca_key_file.name
            crl_kwargs["signing_cert"] = ca_cert_file.name
            # Add list of revoked certificates
            crl_kwargs["revoked"] = revoked
            x509.create_crl(**crl_kwargs)

        # Retrieve serial number from server certificate
        server_cert_details = x509.read_certificate(server_cert_file.name)
        serial_number = server_cert_details["Serial Number"].replace(":", "")
        serial_number = salt.utils.stringutils.to_str(serial_number)

        # Retrieve CRL as text
        crl = M2Crypto.X509.load_crl(ca_crl_file.name).as_text()

        # Cleanup
        os.remove(ca_key_file.name)
        os.remove(ca_cert_file.name)
        os.remove(ca_crl_file.name)
        os.remove(server_cert_file.name)

        # Ensure that the correct server cert serial is amongst
        # the revoked certificates
        self.assertIn(serial_number, crl)

    @pytest.mark.skipif(
        not HAS_M2CRYPTO, reason="Skipping, reason=M2Crypto is unavailable"
    )
    def test_read_certificate(self):
        """
        :return:
        """
        cet = dedent(
            """
                -----BEGIN CERTIFICATE-----
        MIICdDCCAd2gAwIBAgIUH6g+PC0bGKSY4LMq7PISP09M5B4wDQYJKoZIhvcNAQEL
        BQAwTDELMAkGA1UEBhMCVVMxEDAOBgNVBAgMB0FyaXpvbmExEzARBgNVBAcMClNj
        b3R0c2RhbGUxFjAUBgNVBAoMDVN1cGVyIFdpZGdpdHMwHhcNMjEwMzIzMDExNDE2
        WhcNMjIwMzIzMDExNDE2WjBMMQswCQYDVQQGEwJVUzEQMA4GA1UECAwHQXJpem9u
        YTETMBEGA1UEBwwKU2NvdHRzZGFsZTEWMBQGA1UECgwNU3VwZXIgV2lkZ2l0czCB
        nzANBgkqhkiG9w0BAQEFAAOBjQAwgYkCgYEAvtFFZP47UkzyAmVWtBnVHuXwe7iK
        yu19c3qx59KPVAMHkMKgCew4S2KBMDHySBVnspiEz1peP1ywozcP1tIeWHG6aY/7
        j2ewzl5bJ4HZPDBnEOYzGsC/NM8YY3qFlrteda/awvwoF99MkpVlrcLBMJzjt/c8
        HjuBb0zTlnm4r7ECAwEAAaNTMFEwHQYDVR0OBBYEFJwdb0PKsvu3dU0j3kx3uP4B
        NGpfMB8GA1UdIwQYMBaAFJwdb0PKsvu3dU0j3kx3uP4BNGpfMA8GA1UdEwEB/wQF
        MAMBAf8wDQYJKoZIhvcNAQELBQADgYEAZblVv70rSk6+7ti3mYxVo48VLf3hG5R/
        rMd434WYTeDOWlvl5GSklrBc4ToBW5GsJe/+JaFbUFo9YB+a0K0xjyNZ5CWWiaxg
        3lwqTx6vwK1ucS18B+nt2qqyq9hL0UvpSB7gH4KeCwCMDIfRMsrPi32jg1RyKftD
        B+O0S5LeuJw=
        -----END CERTIFICATE-----
        """
        )
        ret = x509.read_certificate(cet)
        assert "MD5 Finger Print" in ret


class X509FipsTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {x509: {"__opts__": {"fips_mode": True}}}

    @pytest.mark.skipif(
        not HAS_M2CRYPTO, reason="Skipping, reason=M2Crypto is unavailable"
    )
    def test_read_certificate(self):
        """
        :return:
        """
        cet = dedent(
            """
                -----BEGIN CERTIFICATE-----
        MIICdDCCAd2gAwIBAgIUH6g+PC0bGKSY4LMq7PISP09M5B4wDQYJKoZIhvcNAQEL
        BQAwTDELMAkGA1UEBhMCVVMxEDAOBgNVBAgMB0FyaXpvbmExEzARBgNVBAcMClNj
        b3R0c2RhbGUxFjAUBgNVBAoMDVN1cGVyIFdpZGdpdHMwHhcNMjEwMzIzMDExNDE2
        WhcNMjIwMzIzMDExNDE2WjBMMQswCQYDVQQGEwJVUzEQMA4GA1UECAwHQXJpem9u
        YTETMBEGA1UEBwwKU2NvdHRzZGFsZTEWMBQGA1UECgwNU3VwZXIgV2lkZ2l0czCB
        nzANBgkqhkiG9w0BAQEFAAOBjQAwgYkCgYEAvtFFZP47UkzyAmVWtBnVHuXwe7iK
        yu19c3qx59KPVAMHkMKgCew4S2KBMDHySBVnspiEz1peP1ywozcP1tIeWHG6aY/7
        j2ewzl5bJ4HZPDBnEOYzGsC/NM8YY3qFlrteda/awvwoF99MkpVlrcLBMJzjt/c8
        HjuBb0zTlnm4r7ECAwEAAaNTMFEwHQYDVR0OBBYEFJwdb0PKsvu3dU0j3kx3uP4B
        NGpfMB8GA1UdIwQYMBaAFJwdb0PKsvu3dU0j3kx3uP4BNGpfMA8GA1UdEwEB/wQF
        MAMBAf8wDQYJKoZIhvcNAQELBQADgYEAZblVv70rSk6+7ti3mYxVo48VLf3hG5R/
        rMd434WYTeDOWlvl5GSklrBc4ToBW5GsJe/+JaFbUFo9YB+a0K0xjyNZ5CWWiaxg
        3lwqTx6vwK1ucS18B+nt2qqyq9hL0UvpSB7gH4KeCwCMDIfRMsrPi32jg1RyKftD
        B+O0S5LeuJw=
        -----END CERTIFICATE-----
        """
        )
        ret = x509.read_certificate(cet)
        assert "MD5 Finger Print" not in ret



"""
Integration tests for the jinja includes in states
"""

import logging

import pytest

log = logging.getLogger(__name__)


@pytest.mark.slow_test
def test_issue_64111(salt_master, salt_minion, salt_call_cli):
    # This needs to be an integration test. A functional test does not trigger
    # the issue fixed.

    macros_jinja = """
    {% macro a_jinja_macro(arg) -%}
    {{ arg }}
    {%- endmacro %}
    """

    init_sls = """
    include:
      - common.file1
    """

    file1_sls = """
    {% from 'common/macros.jinja' import a_jinja_macro with context %}

    a state id:
      cmd.run:
        - name: echo {{ a_jinja_macro("hello world") }}
    """
    tf = salt_master.state_tree.base.temp_file

    with tf("common/macros.jinja", macros_jinja):
        with tf("common/init.sls", init_sls):
            with tf("common/file1.sls", file1_sls):
                ret = salt_call_cli.run("state.apply", "common")
                assert ret.returncode == 0



import importlib
import subprocess
import types

import pytest

import salt.client.ssh.shell as shell
from tests.support.mock import MagicMock, PropertyMock, patch


@pytest.fixture
def keys(tmp_path):
    pub_key = tmp_path / "ssh" / "testkey.pub"
    priv_key = tmp_path / "ssh" / "testkey"
    return types.SimpleNamespace(pub_key=pub_key, priv_key=priv_key)


@pytest.mark.skip_on_windows(reason="Windows does not support salt-ssh")
@pytest.mark.skip_if_binaries_missing("ssh", "ssh-keygen", check_all=True)
def test_ssh_shell_key_gen(keys):
    """
    Test ssh key_gen
    """
    shell.gen_key(str(keys.priv_key))
    assert keys.priv_key.exists()
    assert keys.pub_key.exists()
    # verify there is not a passphrase set on key
    ret = subprocess.check_output(
        ["ssh-keygen", "-f", str(keys.priv_key), "-y"],
        timeout=30,
    )
    assert ret.decode().startswith("ssh-rsa")


@pytest.mark.skip_on_windows(reason="Windows does not support salt-ssh")
@pytest.mark.skip_if_binaries_missing("ssh", "ssh-keygen", check_all=True)
def test_ssh_shell_exec_cmd(caplog):
    """
    Test executing a command and ensuring the password
    is not in the stdout/stderr logs.
    """
    passwd = "12345"
    opts = {"_ssh_version": (4, 9)}
    host = ""
    _shell = shell.Shell(opts=opts, host=host)
    _shell.passwd = passwd
    with patch.object(_shell, "_split_cmd", return_value=["echo", passwd]):
        ret = _shell.exec_cmd(f"echo {passwd}")
        assert not any([x for x in ret if passwd in str(x)])
        assert passwd not in caplog.text

    with patch.object(_shell, "_split_cmd", return_value=["ls", passwd]):
        ret = _shell.exec_cmd(f"ls {passwd}")
        assert not any([x for x in ret if passwd in str(x)])
        assert passwd not in caplog.text


@pytest.mark.parametrize(
    "text, sanitize, expected",
    [
        ("-oServerAliveInterval=60", "Server", "-oServerAliveInterval=60"),
        (
            "-o ServerAliveInterval=60 --password Server",
            "Server",
            "-o ServerAliveInterval=60 --password ******",
        ),
    ],
)
def test_ssh_shell_sanitize(text, sanitize, expected):
    """
    Test that _sanitize_str doesn't replace strings inside of other words.
    """
    shl = shell.Shell({}, "localhost")
    res = shl._sanitize_str(text, sanitize)

    assert res == expected


def test_run_cmd_password_prompt():
    """
    When using a password that has the same value as the shell
    buffer, test that the sanitization is done after internal
    matching, e.g. with "SSH_PRIVATE_KEY_PASSWORD_PROMPT_RE".
    """
    passwd = "password"
    shl = shell.Shell({}, "localhost", passwd=passwd)
    mock_ssh_re = MagicMock()

    mock_term = MagicMock()
    mock_term.recv.return_value = (passwd, None)

    with patch.object(shell, "SSH_PRIVATE_KEY_PASSWORD_PROMPT_RE", mock_ssh_re), patch(
        "salt.utils.vt.Terminal", return_value=mock_term
    ):
        shl._run_cmd("test_cmd")

    mock_ssh_re.search.assert_called_once_with(passwd)


def test_ssh_shell_exec_cmd_waits_for_term_close_before_reading_exit_status():
    """
    Ensure that the terminal is always closed before accessing its exitstatus.
    """
    term = MagicMock()
    has_unread_data = PropertyMock(side_effect=(True, True, False))
    exitstatus = PropertyMock(
        side_effect=lambda *args: 0 if term._closed is True else None
    )
    term.close.side_effect = lambda *args, **kwargs: setattr(term, "_closed", True)
    type(term).has_unread_data = has_unread_data
    type(term).exitstatus = exitstatus
    term.recv.side_effect = (("hi ", ""), ("there", ""), (None, None), (None, None))
    shl = shell.Shell({}, "localhost")
    with patch("salt.utils.vt.Terminal", autospec=True, return_value=term):
        stdout, stderr, retcode = shl.exec_cmd("do something")
    assert stdout == "hi there"
    assert stderr == ""
    assert retcode == 0


def test_ssh_shell_exec_cmd_returns_status_code_with_highest_bit_set_if_process_dies():
    """
    Ensure that if a child process dies as the result of a signal instead of exiting
    regularly, the shell returns the signal code encoded in the lowest seven bits with
    the highest one set, not None.
    """
    term = MagicMock()
    term.exitstatus = None
    term.signalstatus = 9
    has_unread_data = PropertyMock(side_effect=(True, True, False))
    type(term).has_unread_data = has_unread_data
    term.recv.side_effect = (
        ("", "leave me alone"),
        ("", " please"),
        (None, None),
        (None, None),
    )
    shl = shell.Shell({}, "localhost")
    with patch("salt.utils.vt.Terminal", autospec=True, return_value=term):
        stdout, stderr, retcode = shl.exec_cmd("do something")
    assert stdout == ""
    assert stderr == "leave me alone please"
    assert retcode == 137


@pytest.fixture()
def mock_bin_paths():
    """Automatically apply fixture to all tests that need it."""
    with patch("salt.utils.path.which") as mock_which:
        mock_which.side_effect = lambda x: {
            "ssh-keygen": "/custom/ssh-keygen",
            "ssh": "/custom/ssh",
            "scp": "/custom/scp",
        }.get(x, None)
        importlib.reload(shell)
        yield
    importlib.reload(shell)


def test_gen_key_uses_custom_ssh_keygen_path(mock_bin_paths):
    """Test that gen_key function uses the correct ssh-keygen path."""
    with patch("subprocess.call") as mock_call:
        shell.gen_key("/dev/null")

        # Extract the first argument of the first call to subprocess.call
        args, _ = mock_call.call_args

        # Assert that the first part of the command is the custom ssh-keygen path
        assert args[0][0] == "/custom/ssh-keygen"


def test_ssh_command_execution_uses_custom_path(mock_bin_paths):
    options = {"_ssh_version": (4, 9)}
    _shell = shell.Shell(opts=options, host="example.com")
    cmd_string = _shell._cmd_str("ls -la")
    assert "/custom/ssh" in cmd_string


def test_scp_command_execution_uses_custom_path(mock_bin_paths):
    _shell = shell.Shell(opts={}, host="example.com")
    with patch.object(
        _shell, "_run_cmd", return_value=(None, None, None)
    ) as mock_run_cmd:
        _shell.send("source_file.txt", "/path/dest_file.txt")
        # The command string passed to _run_cmd should include the custom scp path
        args, _ = mock_run_cmd.call_args
        assert "/custom/scp" in args[0]
        assert "source_file.txt example.com:/path/dest_file.txt" in args[0]



import logging
import os
import pathlib

import pytest

import salt.states.file as filestate
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {filestate: {"__salt__": {}, "__opts__": {}, "__env__": "base"}}


def test__gen_recurse_managed_files():
    """
    Test _gen_recurse_managed_files to make sure it puts symlinks at the end of the list of files.
    """
    target_dir = pathlib.Path(f"{os.sep}some{os.sep}path{os.sep}target")
    cp_list_master = MagicMock(
        return_value=[
            "target/symlink",
            "target/just_a_file.txt",
            "target/not_a_symlink/symlink",
            "target/notasymlink",
        ],
    )
    cp_list_master_symlinks = MagicMock(
        return_value={
            "target/symlink": f"{target_dir}{os.sep}not_a_symlink{os.sep}symlink"
        }
    )
    patch_salt = {
        "cp.list_master": cp_list_master,
        "cp.list_master_symlinks": cp_list_master_symlinks,
    }
    with patch.dict(filestate.__salt__, patch_salt):
        files, dirs, links, keep = filestate._gen_recurse_managed_files(
            name=str(target_dir), source=f"salt://{target_dir.name}", keep_symlinks=True
        )
    expected = (
        f"{os.sep}some{os.sep}path{os.sep}target{os.sep}symlink",
        "salt://target/symlink?saltenv=base",
    )
    assert files[-1] == expected



"""
Watch files and translate the changes into salt events

:depends:   - pyinotify Python module >= 0.9.5

:Caution:   Using generic mask options like open, access, ignored, and
            closed_nowrite with reactors can easily cause the reactor
            to loop on itself. To mitigate this behavior, consider
            setting the `disable_during_state_run` flag to `True` in
            the beacon configuration.

:note: The `inotify` beacon only works on OSes that have `inotify`
       kernel support.

"""

import collections
import fnmatch
import logging
import os
import re

import salt.utils.beacons

try:
    import pyinotify

    HAS_PYINOTIFY = True
    DEFAULT_MASK = pyinotify.IN_CREATE | pyinotify.IN_DELETE | pyinotify.IN_MODIFY
    MASKS = {}
    for var in dir(pyinotify):
        if var.startswith("IN_"):
            key = var[3:].lower()
            MASKS[key] = getattr(pyinotify, var)
except ImportError:
    HAS_PYINOTIFY = False
    DEFAULT_MASK = None

__virtualname__ = "inotify"

log = logging.getLogger(__name__)


def __virtual__():
    if HAS_PYINOTIFY:
        return __virtualname__
    err_msg = "pyinotify library is missing."
    log.error("Unable to load inotify beacon: %s", err_msg)
    return False, err_msg


def _get_mask(mask):
    """
    Return the int that represents the mask
    """
    return MASKS.get(mask, 0)


def _enqueue(revent):
    """
    Enqueue the event
    """
    __context__["inotify.queue"].append(revent)


def _get_notifier(config):
    """
    Check the context for the notifier and construct it if not present
    """
    beacon_name = config.get("_beacon_name", "inotify")
    notifier = f"{beacon_name}.notifier"
    if notifier not in __context__:
        __context__["inotify.queue"] = collections.deque()
        wm = pyinotify.WatchManager()
        __context__[notifier] = pyinotify.Notifier(wm, _enqueue)
        if (
            "coalesce" in config
            and isinstance(config["coalesce"], bool)
            and config["coalesce"]
        ):
            __context__[notifier].coalesce_events()
    return __context__[notifier]


def validate(config):
    """
    Validate the beacon configuration
    """

    VALID_MASK = [
        "access",
        "attrib",
        "close_nowrite",
        "close_write",
        "create",
        "delete",
        "delete_self",
        "excl_unlink",
        "ignored",
        "modify",
        "moved_from",
        "moved_to",
        "move_self",
        "oneshot",
        "onlydir",
        "open",
        "unmount",
    ]

    # Configuration for inotify beacon should be a dict of dicts
    if not isinstance(config, list):
        return False, "Configuration for inotify beacon must be a list."
    else:
        config = salt.utils.beacons.list_to_dict(config)

        if "files" not in config:
            return False, "Configuration for inotify beacon must include files."
        else:
            if not isinstance(config["files"], dict):
                return (
                    False,
                    "Configuration for inotify beacon invalid, files must be a dict.",
                )

            for path in config.get("files"):

                if not isinstance(config["files"][path], dict):
                    return (
                        False,
                        "Configuration for inotify beacon must be a list of"
                        " dictionaries.",
                    )
                else:
                    if not any(
                        j in ["mask", "recurse", "auto_add"]
                        for j in config["files"][path]
                    ):
                        return (
                            False,
                            "Configuration for inotify beacon must contain mask,"
                            " recurse or auto_add items.",
                        )

                    if "auto_add" in config["files"][path]:
                        if not isinstance(config["files"][path]["auto_add"], bool):
                            return (
                                False,
                                "Configuration for inotify beacon auto_add must be"
                                " boolean.",
                            )

                    if "recurse" in config["files"][path]:
                        if not isinstance(config["files"][path]["recurse"], bool):
                            return (
                                False,
                                "Configuration for inotify beacon recurse must be"
                                " boolean.",
                            )

                    if "mask" in config["files"][path]:
                        if not isinstance(config["files"][path]["mask"], list):
                            return (
                                False,
                                "Configuration for inotify beacon mask must be list.",
                            )
                        for mask in config["files"][path]["mask"]:
                            if mask not in VALID_MASK:
                                return (
                                    False,
                                    "Configuration for inotify beacon invalid mask"
                                    " option {}.".format(mask),
                                )
    return True, "Valid beacon configuration"


def beacon(config):
    """
    Watch the configured files

    Example Config

    .. code-block:: yaml

        beacons:
          inotify:
            - files:
                /path/to/file/or/dir:
                  mask:
                    - open
                    - create
                    - close_write
                  recurse: True
                  auto_add: True
                  exclude:
                    - /path/to/file/or/dir/exclude1
                    - /path/to/file/or/dir/exclude2
                    - /path/to/file/or/dir/regex[a-m]*$:
                        regex: True
            - coalesce: True

    The mask list can contain the following events (the default mask is create,
    delete, and modify):

    * access            - File accessed
    * attrib            - File metadata changed
    * close_nowrite     - Unwritable file closed
    * close_write       - Writable file closed
    * create            - File created in watched directory
    * delete            - File deleted from watched directory
    * delete_self       - Watched file or directory deleted
    * modify            - File modified
    * moved_from        - File moved out of watched directory
    * moved_to          - File moved into watched directory
    * move_self         - Watched file moved
    * open              - File opened

    The mask can also contain the following options:

    * dont_follow       - Don't dereference symbolic links
    * excl_unlink       - Omit events for children after they have been unlinked
    * oneshot           - Remove watch after one event
    * onlydir           - Operate only if name is directory

    recurse:
      Recursively watch files in the directory
    auto_add:
      Automatically start watching files that are created in the watched directory
    exclude:
      Exclude directories or files from triggering events in the watched directory.
      Can use regex if regex is set to True
    coalesce:
      If this coalescing option is enabled, events are filtered based on
      their unicity, only unique events are enqueued, doublons are discarded.
      An event is unique when the combination of its fields (wd, mask,
      cookie, name) is unique among events of a same batch. After a batch of
      events is processed any events are accepted again.
      This option is top-level (at the same level as the path) and therefore
      affects all paths that are being watched. This is due to this option
      being at the Notifier level in pyinotify.
    """

    whitelist = ["_beacon_name"]
    config = salt.utils.beacons.remove_hidden_options(config, whitelist)

    config = salt.utils.beacons.list_to_dict(config)

    ret = []
    notifier = _get_notifier(config)
    wm = notifier._watch_manager

    # Read in existing events
    if notifier.check_events(1):
        notifier.read_events()
        notifier.process_events()
        queue = __context__["inotify.queue"]
        while queue:
            event = queue.popleft()

            _append = True
            # Find the matching path in config
            path = event.path
            while path != "/":
                if path in config.get("files", {}):
                    break
                path = os.path.dirname(path)

            excludes = config["files"].get(path, {}).get("exclude", "")

            if excludes and isinstance(excludes, list):
                for exclude in excludes:
                    if isinstance(exclude, dict):
                        _exclude = next(iter(exclude))
                        if exclude[_exclude].get("regex", False):
                            try:
                                if re.search(_exclude, event.pathname):
                                    _append = False
                            except Exception:  # pylint: disable=broad-except
                                log.warning("Failed to compile regex: %s", _exclude)
                        else:
                            exclude = _exclude
                    elif "*" in exclude:
                        if fnmatch.fnmatch(event.pathname, exclude):
                            _append = False
                    else:
                        if event.pathname.startswith(exclude):
                            _append = False

            if _append:
                sub = {
                    "tag": event.path,
                    "path": event.pathname,
                    "change": event.maskname,
                }
                ret.append(sub)
            else:
                log.info("Excluding %s from event for %s", event.pathname, path)

    # Get paths currently being watched
    current = set()
    for wd in wm.watches:
        current.add(wm.watches[wd].path)

    # Update existing watches and add new ones
    # TODO: make the config handle more options
    for path in config.get("files", ()):

        if isinstance(config["files"][path], dict):
            mask = config["files"][path].get("mask", DEFAULT_MASK)
            if isinstance(mask, list):
                r_mask = 0
                for sub in mask:
                    r_mask |= _get_mask(sub)
            elif isinstance(mask, bytes):
                r_mask = _get_mask(mask)
            else:
                r_mask = mask
            mask = r_mask
            rec = config["files"][path].get("recurse", False)
            auto_add = config["files"][path].get("auto_add", False)
        else:
            mask = DEFAULT_MASK
            rec = False
            auto_add = False

        if path in current:
            for wd in wm.watches:
                if path == wm.watches[wd].path:
                    update = False
                    if wm.watches[wd].mask != mask:
                        update = True
                    if wm.watches[wd].auto_add != auto_add:
                        update = True
                    if update:
                        wm.update_watch(wd, mask=mask, rec=rec, auto_add=auto_add)
        elif os.path.exists(path):
            excludes = config["files"][path].get("exclude", "")
            excl = None
            if isinstance(excludes, list):
                excl = []
                for exclude in excludes:
                    if isinstance(exclude, dict):
                        excl.append(list(exclude)[0])
                    else:
                        excl.append(exclude)
                excl = pyinotify.ExcludeFilter(excl)

            wm.add_watch(path, mask, rec=rec, auto_add=auto_add, exclude_filter=excl)

    # Return event data
    return ret


def close(config):
    config = salt.utils.beacons.list_to_dict(config)
    beacon_name = config.get("_beacon_name", "inotify")
    notifier = f"{beacon_name}.notifier"
    if notifier in __context__:
        __context__[notifier].stop()
        del __context__[notifier]



"""
    :codeauthor: Eric Radman <ericshane@eradman.com>
"""

import pytest

from salt.cloud.clouds import joyent
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.skipif(
        joyent.HAS_REQUIRED_CRYPTO is False,
        reason="PyCrypto or Cryptodome not installed",
    )
]


def _fake_wait_for_ip(
    check_for_ip_fn, interval=None, timeout=None, interval_multiplier=None
):
    """
    Callback that returns immediately instead of waiting
    """
    assert isinstance(interval, int)
    assert isinstance(timeout, int)
    assert isinstance(interval_multiplier, int)
    return check_for_ip_fn()


@pytest.fixture
def configure_loader_modules():
    with patch("salt.utils.cloud.wait_for_ip", _fake_wait_for_ip):
        yield {
            joyent: {
                "__utils__": {
                    "cloud.fire_event": MagicMock(),
                    "cloud.bootstrap": MagicMock(),
                },
                "__opts__": {
                    "sock_dir": True,
                    "transport": True,
                    "providers": {"my_joyent": {}},
                    "profiles": {"my_joyent": {}},
                },
                "__active_provider_name__": "my_joyent:joyent",
            }
        }


@pytest.fixture
def vm_():
    return {
        "profile": "my_joyent",
        "name": "vm3",
        "driver": "joyent",
        "size": "k4-highcpu-kvm-750M",
        "image": "freebsd10",
        "location": "us-east-1",
    }


def test_query_instance_init(vm_):
    """
    Initial provisioning, no IP assigned
    """
    # Not yet reachable
    reply = (200, {"state": "provisioning"})
    with patch.object(joyent, "show_instance", return_value=reply):
        result = joyent.query_instance(vm_)
    joyent.__utils__["cloud.fire_event"].assert_called_once()
    assert result is None


def test_query_instance_has_ip(vm_):
    """
    IP address assigned but not yet ready
    """
    reply = (200, {"primaryIp": "1.1.1.1", "state": "provisioning"})
    with patch.object(joyent, "show_instance", return_value=reply):
        result = joyent.query_instance(vm_)
    joyent.__utils__["cloud.fire_event"].assert_called_once()
    assert result is None


def test_query_instance_ready(vm_):
    """
    IP address assigned, and VM is ready
    """
    reply = (200, {"primaryIp": "1.1.1.1", "state": "running"})
    with patch.object(joyent, "show_instance", return_value=reply):
        result = joyent.query_instance(vm_)
    joyent.__utils__["cloud.fire_event"].assert_called_once()
    assert result == "1.1.1.1"



"""
Tests for the file state
"""

import os

import pytest

import salt.utils.win_dacl


@pytest.fixture
def setup_symlink(tmp_path):
    name = tmp_path / "test-symlink"
    name.mkdir()
    if salt.utils.platform.is_windows():
        principal = salt.utils.win_functions.get_current_user()
        salt.utils.win_dacl.set_owner(obj_name=str(name), principal=principal)
        salt.utils.win_dacl.set_inheritance(obj_name=str(name), enabled=True)
    source_file = name / "source"
    source_file.write_text("jaguar", encoding="utf-8")
    assert source_file.exists()
    assert source_file.is_file()
    assert name.exists()
    assert name.is_dir()
    return str(name), str(source_file)


@pytest.fixture(params=["str", "list"])
def sls_content(setup_symlink, request):
    name, symlink_file = setup_symlink
    string_contents = """
    {name}/testing:
      file.symlink:
        - target: {symlink_file}
        - makedirs: true
        - check_cmd: grep 'jaguar' {symlink_file}
    """.format(
        name=name, symlink_file=symlink_file
    )
    list_contents = """
    {name}/testing:
      file.symlink:
        - target: {symlink_file}
        - makedirs: true
        - check_cmd:
          - grep 'jaguar' {symlink_file}
          - grep "j" {symlink_file}
    """.format(
        name=name, symlink_file=symlink_file
    )
    if request.param == str:
        return string_contents, name, symlink_file
    else:
        return list_contents, name, symlink_file


@pytest.mark.parametrize("verify_ssl", [True, False])
@pytest.mark.slow_test
def test_get_source_sum_verify_ssl_false(
    salt_call_cli, tmp_path, ssl_webserver, verify_ssl, this_txt_file
):
    """
    test verify_ssl with get_source_sum
    """
    web_file = ssl_webserver.url("this.txt")
    if verify_ssl:
        # Clean the cached this.txt if it exists as it will fail the test
        # because it won't fetch it again
        ret = salt_call_cli.run("cp.is_cached", web_file, saltenv="base")
        assert ret.returncode == 0
        if ret.data:
            os.unlink(ret.data)
    ret = salt_call_cli.run(
        "--local",
        "file.get_source_sum",
        str(tmp_path / "test_source_sum.txt"),
        source=web_file,
        source_hash=web_file + ".sha256",
        verify_ssl=verify_ssl,
    )
    if not verify_ssl:
        assert ret.data["hsum"] == this_txt_file.sha256
        assert ret.data["hash_type"] == "sha256"
    else:
        assert "SSL: CERTIFICATE_VERIFY_FAILED" in ret.stderr


@pytest.mark.parametrize("verify_ssl", [True, False])
@pytest.mark.slow_test
def test_get_managed_verify_ssl(salt_call_cli, tmp_path, ssl_webserver, verify_ssl):
    """
    test verify_ssl with get_managed
    """
    web_file = ssl_webserver.url("this.txt")
    if verify_ssl:
        # Clean the cached this.txt if it exists as it will fail the test
        # because it won't fetch it again
        ret = salt_call_cli.run("cp.is_cached", web_file, saltenv="base")
        assert ret.returncode == 0
        if ret.data:
            os.unlink(ret.data)
    ret = salt_call_cli.run(
        "--local",
        "file.get_managed",
        str(tmp_path / "test_managed.txt"),
        template="",
        source=web_file,
        source_hash=web_file + ".sha256",
        source_hash_name="",
        user="",
        group="",
        mode="",
        attrs="",
        saltenv="base",
        context={},
        defaults="",
        skip_verify=True,
        verify_ssl=verify_ssl,
    )
    if not verify_ssl:
        assert "this.txt" in ret.data[0]
    else:
        assert "SSL: CERTIFICATE_VERIFY_FAILED" in ret.stdout


@pytest.mark.parametrize("verify_ssl", [True, False])
@pytest.mark.slow_test
def test_manage_file_verify_ssl(
    salt_call_cli, tmp_path, ssl_webserver, verify_ssl, this_txt_file
):
    """
    test verify_ssl with manage_file
    """
    test_file = tmp_path / "test_manage_file.txt"
    web_file = ssl_webserver.url("this.txt")
    if verify_ssl:
        # Clean the cached this.txt if it exists as it will fail the test
        # because it won't fetch it again
        ret = salt_call_cli.run("cp.is_cached", web_file, saltenv="base")
        assert ret.returncode == 0
        if ret.data:
            os.unlink(ret.data)
    ret = salt_call_cli.run(
        "--local",
        "file.manage_file",
        str(test_file),
        sfn="",
        ret="",
        source=web_file,
        source_sum={"hash_type": "sha256", "hsum": this_txt_file.sha256},
        user="",
        group="",
        mode="",
        attrs="",
        saltenv="base",
        backup="",
        verify_ssl=verify_ssl,
    )
    if not verify_ssl:
        assert ret.data["changes"] == {"diff": "New file", "mode": "0000"}
        assert ret.data["comment"] == f"File {test_file} updated"
    else:
        assert "SSL: CERTIFICATE_VERIFY_FAILED" in ret.stderr


@pytest.mark.parametrize("verify_ssl", [True, False])
@pytest.mark.slow_test
def test_check_managed_changes_verify_ssl(
    salt_call_cli, tmp_path, ssl_webserver, verify_ssl
):
    """
    test verify_ssl with check_managed_changes
    """
    test_file = tmp_path / "test_managed_changes.txt"
    web_url = ssl_webserver.url("this.txt")
    if verify_ssl:
        # Clean the cached this.txt if it exists as it will fail the test
        # because it won't fetch it again
        ret = salt_call_cli.run("cp.is_cached", web_url, saltenv="base")
        assert ret.returncode == 0
        if ret.data:
            os.unlink(ret.data)
    ret = salt_call_cli.run(
        "--local",
        "file.check_managed_changes",
        str(test_file),
        source=web_url,
        source_hash=web_url + ".sha256",
        source_hash_name="",
        user="",
        group="",
        mode="",
        attrs="",
        template="jinja",
        context="",
        defaults="",
        saltenv="base",
        verify_ssl=verify_ssl,
    )

    if not verify_ssl:
        assert ret.data["newfile"] == str(test_file)
    else:
        assert "SSL: CERTIFICATE_VERIFY_FAILED" in ret.stderr


@pytest.mark.parametrize("verify_ssl", [True, False])
@pytest.mark.slow_test
def test_check_file_meta_verify_ssl(
    salt_call_cli, tmp_path, ssl_webserver, verify_ssl, this_txt_file
):
    """
    test verify_ssl with check_file_meta
    """
    test_file = tmp_path / "test_check_file_meta.txt"
    test_file.write_text("test check_file_meta")
    web_url = ssl_webserver.url("this.txt")
    if verify_ssl:
        # Clean the cached this.txt if it exists as it will fail the test
        # because it won't fetch it again
        ret = salt_call_cli.run("cp.is_cached", web_url, saltenv="base")
        assert ret.returncode == 0
        if ret.data:
            os.unlink(ret.data)
    ret = salt_call_cli.run(
        "--local",
        "file.check_file_meta",
        str(test_file),
        sfn="",
        source=web_url,
        source_sum={"hash_type": "sha256", "hsum": this_txt_file.sha256},
        user="",
        mode="",
        group="",
        attrs="",
        saltenv="base",
        verify_ssl=verify_ssl,
    )

    if not verify_ssl:
        assert (
            len([x for x in ["diff", "user", "group", "mode"] if x in ret.data.keys()])
            == 4
        )
    else:
        assert "SSL: CERTIFICATE_VERIFY_FAILED" in ret.stderr


@pytest.mark.slow_test
def test_create_symlink_with_check_cmd(salt_call_cli, salt_master, sls_content):
    """
    file.symlink test to make sure check_cmd runs before
    creating the symlink and can run a list of commands
    """
    sls_contents, name, symlink_file = sls_content
    with salt_master.state_tree.base.temp_file("test_symlink.sls", sls_contents):
        ret = salt_call_cli.run("state.apply", "test_symlink")
        symlink_file = os.path.join(name, "testing")
        assert os.path.exists(symlink_file)
        assert os.path.islink(symlink_file)
        expected_comment = "check_cmd determined the state succeeded"
        assert expected_comment in ret.stdout



"""
Module for listing programs that automatically run on startup
(very alpha...not tested on anything but my Win 7x64)
"""

import os

import salt.utils.platform

# Define a function alias in order not to shadow built-in's
__func_alias__ = {"list_": "list"}

# Define the module's virtual name
__virtualname__ = "autoruns"


def __virtual__():
    """
    Only works on Windows systems
    """

    if salt.utils.platform.is_windows():
        return __virtualname__
    return (False, "Module win_autoruns: module only works on Windows systems")


def _get_dirs(user_dir, startup_dir):
    """
    Return a list of startup dirs
    """
    try:
        users = os.listdir(user_dir)
    except OSError:  # pylint: disable=E0602
        users = []

    full_dirs = []
    for user in users:
        full_dir = os.path.join(user_dir, user, startup_dir)
        if os.path.exists(full_dir):
            full_dirs.append(full_dir)
    return full_dirs


def list_():
    """
    Get a list of automatically running programs

    CLI Example:

    .. code-block:: bash

        salt '*' autoruns.list
    """
    autoruns = {}

    # Find autoruns in registry
    keys = [
        "HKLM\\Software\\Microsoft\\Windows\\CurrentVersion\\Run",
        "HKLM\\Software\\Microsoft\\Windows\\CurrentVersion\\Run /reg:64",
        "HKCU\\Software\\Microsoft\\Windows\\CurrentVersion\\Run",
    ]
    for key in keys:
        autoruns[key] = []
        cmd = ["reg", "query", key]
        for line in __salt__["cmd.run"](cmd, python_shell=False).splitlines():
            if (
                line and line[0:4] != "HKEY" and line[0:5] != "ERROR"
            ):  # Remove junk lines
                autoruns[key].append(line)

    # Find autoruns in user's startup folder
    user_dir = "C:\\Documents and Settings\\"
    startup_dir = "\\Start Menu\\Programs\\Startup"
    full_dirs = _get_dirs(user_dir, startup_dir)
    if not full_dirs:
        user_dir = "C:\\Users\\"
        startup_dir = (
            "\\AppData\\Roaming\\Microsoft\\Windows\\Start Menu\\Programs\\Startup"
        )
        full_dirs = _get_dirs(user_dir, startup_dir)

    for full_dir in full_dirs:
        files = os.listdir(full_dir)
        autoruns[full_dir] = []
        for single_file in files:
            autoruns[full_dir].append(single_file)

    return autoruns



# pylint: disable=W0223,W0231,W0221
import logging

import salt.modules.netmiko_mod as netmiko_mod
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase

log = logging.getLogger(__name__)


class MockNetmikoConnection:
    def __init__(self, *args, **kwargs):
        self.user = kwargs["username"]
        self.password = kwargs["password"]

    def is_alive(self):
        return False

    def send_config_set(self, *args, **kwargs):
        return args, kwargs


def mock_netmiko_args():
    return {"username": "salt", "password": "password"}


def mock_netmiko_conn():
    return MockNetmikoConnection(**mock_netmiko_args())


def mock_file_apply_template_on_contents(*args):
    return args[0]


def mock_prepare_connection(**kwargs):
    return MockNetmikoConnection(**kwargs), {}


class NetmikoTestCase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        return {
            netmiko_mod: {
                "__salt__": {
                    "file.apply_template_on_contents": mock_file_apply_template_on_contents,
                },
                "__proxy__": {
                    "netmiko.conn": mock_netmiko_conn,
                    "netmiko.args": mock_netmiko_args,
                },
                "_prepare_connection": mock_prepare_connection,
            }
        }

    def test_send_config(self):
        """
        Test netmiko.send_config function
        """
        _, ret = netmiko_mod.send_config(
            config_commands=["ls", "echo hello world"],
            config_mode_command="config config-sess",
        )
        self.assertEqual(ret.get("config_commands"), ["ls", "echo hello world"])
        self.assertEqual(ret.get("config_mode_command"), "config config-sess")

        with patch.dict(
            netmiko_mod.__proxy__, {"netmiko.conn": MagicMock(return_value=None)}
        ):
            _, ret = netmiko_mod.send_config(
                config_commands=["ls", "echo hello world"],
                config_mode_command="config config-sess",
            )
            self.assertEqual(ret.get("config_commands"), ["ls", "echo hello world"])
            self.assertEqual(ret.get("config_mode_command"), "config config-sess")

    def test_virtual(self):
        _expected = (
            False,
            "The netmiko execution module requires netmiko library to be installed.",
        )
        with patch("salt.utils.platform.is_proxy", return_value=True, autospec=True):
            with patch.dict(netmiko_mod.__opts__, {"proxy": {"proxytype": "netmiko"}}):
                with patch.object(netmiko_mod, "HAS_NETMIKO", False):
                    ret = netmiko_mod.__virtual__()
                    self.assertTrue(ret)

        _expected = (
            False,
            "The netmiko execution module requires netmiko library to be installed.",
        )
        with patch("salt.utils.platform.is_proxy", return_value=False, autospec=True):
            with patch.dict(netmiko_mod.__opts__, {"proxy": {"proxytype": "esxi"}}):
                with patch.object(netmiko_mod, "HAS_NETMIKO", False):
                    ret = netmiko_mod.__virtual__()
                    self.assertTrue(ret)

        with patch("salt.utils.platform.is_proxy", return_value=True, autospec=True):
            with patch.dict(netmiko_mod.__opts__, {"proxy": {"proxytype": "netmiko"}}):
                with patch.object(netmiko_mod, "HAS_NETMIKO", True):
                    ret = netmiko_mod.__virtual__()
                    self.assertTrue(ret)
                    self.assertEqual(ret, "netmiko")

        with patch("salt.utils.platform.is_proxy", return_value=True, autospec=True):
            with patch.dict(netmiko_mod.__opts__, {"proxy": {"proxytype": "esxi"}}):
                with patch.object(netmiko_mod, "HAS_NETMIKO", True):
                    ret = netmiko_mod.__virtual__()
                    self.assertTrue(ret)
                    self.assertEqual(ret, "netmiko")

        with patch("salt.utils.platform.is_proxy", return_value=False, autospec=True):
            with patch.object(netmiko_mod, "HAS_NETMIKO", True):
                ret = netmiko_mod.__virtual__()
                self.assertEqual(ret, "netmiko")

        with patch("salt.utils.platform.is_proxy", return_value=True, autospec=True):
            with patch.dict(netmiko_mod.__opts__, {"proxy": {"proxytype": "napalm"}}):
                with patch.object(netmiko_mod, "HAS_NETMIKO", True):
                    ret = netmiko_mod.__virtual__()
                    self.assertTrue(ret)
                    self.assertEqual(ret, "netmiko")

        _expected = (False, "Unsupported proxy minion type.")
        with patch("salt.utils.platform.is_proxy", return_value=True, autospec=True):
            with patch.dict(
                netmiko_mod.__opts__, {"proxy": {"proxytype": "deltaproxy"}}
            ):
                with patch.object(netmiko_mod, "HAS_NETMIKO", True):
                    ret = netmiko_mod.__virtual__()
                    self.assertEqual(ret, _expected)



"""
    Test cases for salt.states.git

    :codeauthor: Erik Johnson <erik@saltstack.com>
"""

import logging

import pytest

import salt.states.git as git_state  # Don't potentially shadow GitPython
from tests.support.mock import DEFAULT, MagicMock, Mock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {git_state: {"__env__": "base", "__opts__": {"test": False}, "__salt__": {}}}


def test_latest_no_diff_for_bare_repo(tmp_path):
    """
    This test ensures that we don't attempt to diff when cloning a repo
    using either bare=True or mirror=True.
    """
    name = "https://foo.com/bar/baz.git"
    gitdir = str(tmp_path / "refs")
    isdir_mock = MagicMock(side_effect=lambda path: DEFAULT if path != gitdir else True)

    branches = ["foo", "bar", "baz"]
    tags = ["v1.1.0", "v.1.1.1", "v1.2.0"]
    local_head = "b9ef06ab6b7524eb7c27d740dbbd5109c6d75ee4"
    remote_head = "eef672c1ec9b8e613905dbcd22a4612e31162807"

    git_diff = Mock()
    dunder_salt = {
        "git.current_branch": MagicMock(return_value=branches[0]),
        "git.config_get_regexp": MagicMock(return_value={}),
        "git.diff": git_diff,
        "git.fetch": MagicMock(return_value={}),
        "git.is_worktree": MagicMock(return_value=False),
        "git.list_branches": MagicMock(return_value=branches),
        "git.list_tags": MagicMock(return_value=tags),
        "git.remote_refs": MagicMock(return_value={"HEAD": remote_head}),
        "git.remotes": MagicMock(
            return_value={"origin": {"fetch": name, "push": name}}
        ),
        "git.rev_parse": MagicMock(side_effect=git_state.CommandExecutionError()),
        "git.revision": MagicMock(return_value=local_head),
        "git.version": MagicMock(return_value="1.8.3.1"),
    }
    with patch("os.path.isdir", isdir_mock), patch.dict(
        git_state.__salt__, dunder_salt
    ):
        result = git_state.latest(
            name=name,
            target=tmp_path,
            mirror=True,  # mirror=True implies bare=True
        )
        assert result["result"] is True, result
        git_diff.assert_not_called()


def test_latest_without_target():
    """
    Test latest when called without passing target
    """
    name = "https://foo.com/bar/baz.git"
    with pytest.raises(TypeError):
        # pylint: disable=E1120
        git_state.latest(name)
        # pylint: enable=E1120


def test_detached_without_target():
    """
    Test detached when called without passing target
    """
    name = "https://foo.com/bar/baz.git"
    with pytest.raises(TypeError):
        # pylint: disable=E1120
        git_state.detached(name)
        # pylint: enable=E1120


def test_cloned_without_target():
    """
    Test cloned when called without passing target
    """
    name = "https://foo.com/bar/baz.git"
    with pytest.raises(TypeError):
        # pylint: disable=E1120
        git_state.cloned(name)
        # pylint: enable=E1120



"""
unit tests for the slack engine
"""

import pytest

import salt.engines.slack_bolt_engine as slack_bolt_engine
from tests.support.mock import MagicMock, call, patch

pytestmark = [
    pytest.mark.skipif(
        slack_bolt_engine.HAS_SLACKBOLT is False,
        reason="The slack_bolt is not installed",
    )
]


class MockRunnerClient:
    """
    Mock RunnerClient class
    """

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def asynchronous(self, *args, **kwargs):
        """
        Mock asynchronous method
        """
        return True


class MockLocalClient:
    """
    Mock RunnerClient class
    """

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        pass

    def cmd_async(self, *args, **kwargs):
        """
        Mock cmd_async method
        """
        return True


class MockSlackBoltSocketMode:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def connect(self, *args, **kwargs):
        return True


class MockSlackBoltApp:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

        self.client = MockSlackBoltAppClient()
        self.logger = None
        self.proxy = None

    def message(self, *args, **kwargs):
        return MagicMock(return_value=True)


class MockSlackBoltAppClient:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def chat_postMessage(self, *args, **kwargs):
        return MagicMock(return_value=True)

    def files_upload(self, *args, **kwargs):
        return MagicMock(return_value=True)


@pytest.fixture
def configure_loader_modules():
    return {slack_bolt_engine: {}}


@pytest.fixture
def slack_client(minion_opts):
    app_token = "xapp-x-xxxxxxxxxxx-xxxxxxxxxxxxx-xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
    bot_token = "xoxb-xxxxxxxxxx-xxxxxxxxxxxxxxxxxxxxxxxx"
    trigger = "!"

    with patch.dict(slack_bolt_engine.__opts__, minion_opts):
        with patch(
            "slack_bolt.App", MagicMock(autospec=True, return_value=MockSlackBoltApp())
        ):
            with patch(
                "slack_bolt.adapter.socket_mode.SocketModeHandler",
                MagicMock(autospec=True, return_value=MockSlackBoltSocketMode()),
            ):
                slack_client = slack_bolt_engine.SlackClient(
                    app_token, bot_token, trigger
                )
                yield slack_client


def test_control_message_target(slack_client):
    """
    Test slack engine: control_message_target
    """
    trigger_string = "!"

    loaded_groups = {
        "default": {
            "targets": {},
            "commands": {"cmd.run", "test.ping"},
            "default_target": {"tgt_type": "glob", "target": "*"},
            "users": {"gareth"},
            "aliases": {
                "whoami": {"cmd": "cmd.run whoami"},
                "list_pillar": {"cmd": "pillar.items"},
            },
        }
    }

    slack_user_name = "gareth"

    # Check for correct cmdline
    _expected = (True, {"tgt_type": "glob", "target": "*"}, ["cmd.run", "whoami"])
    text = "!cmd.run whoami"
    target_commandline = slack_client.control_message_target(
        slack_user_name, text, loaded_groups, trigger_string
    )

    assert target_commandline == _expected

    # Check aliases result in correct cmdline
    text = "!whoami"
    target_commandline = slack_client.control_message_target(
        slack_user_name, text, loaded_groups, trigger_string
    )

    assert target_commandline == _expected

    # Check pillar is overridden
    _expected = (
        True,
        {"tgt_type": "glob", "target": "*"},
        ["pillar.items", 'pillar={"hello": "world"}'],
    )
    text = r"""!list_pillar pillar='{"hello": "world"}'"""
    target_commandline = slack_client.control_message_target(
        slack_user_name, text, loaded_groups, trigger_string
    )

    assert target_commandline == _expected

    # Check target is overridden
    _expected = (
        True,
        {"tgt_type": "glob", "target": "localhost"},
        ["cmd.run", "whoami"],
    )
    text = "!cmd.run whoami target='localhost'"
    target_commandline = slack_client.control_message_target(
        slack_user_name, text, loaded_groups, trigger_string
    )

    assert target_commandline == _expected


def test_run_commands_from_slack_async(slack_client):
    """
    Test slack engine: test_run_commands_from_slack_async
    """

    mock_job_status = {
        "20221027001127600438": {
            "data": {"minion": {"return": True, "retcode": 0, "success": True}},
            "function": "test.ping",
        }
    }

    message_generator = [
        {
            "message_data": {
                "client_msg_id": "c1d0c13d-5e78-431e-9921-4786a7d27543",
                "type": "message",
                "text": '!test.ping target="minion"',
                "user": "U02QY11UJ",
                "ts": "1666829486.542159",
                "blocks": [
                    {
                        "type": "rich_text",
                        "block_id": "2vdy",
                        "elements": [
                            {
                                "type": "rich_text_section",
                                "elements": [
                                    {
                                        "type": "text",
                                        "text": '!test.ping target="minion"',
                                    }
                                ],
                            }
                        ],
                    }
                ],
                "team": "T02QY11UG",
                "channel": "C02QY11UQ",
                "event_ts": "1666829486.542159",
                "channel_type": "channel",
            },
            "channel": "C02QY11UQ",
            "user": "U02QY11UJ",
            "user_name": "garethgreenaway",
            "cmdline": ["test.ping"],
            "target": {"target": "minion", "tgt_type": "glob"},
        }
    ]

    mock_files_upload_resp = {
        "ok": True,
        "file": {
            "id": "F047YTDGJF9",
            "created": 1666883749,
            "timestamp": 1666883749,
            "name": "salt-results-20221027081549173603.yaml",
            "title": "salt-results-20221027081549173603",
            "mimetype": "text/plain",
            "filetype": "yaml",
            "pretty_type": "YAML",
            "user": "U0485K894PN",
            "user_team": "T02QY11UG",
            "editable": True,
            "size": 18,
            "mode": "snippet",
            "is_external": False,
            "external_type": "",
            "is_public": True,
            "public_url_shared": False,
            "display_as_bot": False,
            "username": "",
            "url_private": "",
            "url_private_download": "",
            "permalink": "",
            "permalink_public": "",
            "edit_link": "",
            "preview": "minion:\n    True",
            "preview_highlight": "",
            "lines": 2,
            "lines_more": 0,
            "preview_is_truncated": False,
            "comments_count": 0,
            "is_starred": False,
            "shares": {
                "public": {
                    "C02QY11UQ": [
                        {
                            "reply_users": [],
                            "reply_users_count": 0,
                            "reply_count": 0,
                            "ts": "1666883749.485979",
                            "channel_name": "general",
                            "team_id": "T02QY11UG",
                            "share_user_id": "U0485K894PN",
                        }
                    ]
                }
            },
            "channels": ["C02QY11UQ"],
            "groups": [],
            "ims": [],
            "has_rich_preview": False,
            "file_access": "visible",
        },
    }

    patch_app_client_files_upload = patch.object(
        MockSlackBoltAppClient,
        "files_upload",
        MagicMock(autospec=True, return_value=mock_files_upload_resp),
    )
    patch_app_client_chat_postMessage = patch.object(
        MockSlackBoltAppClient,
        "chat_postMessage",
        MagicMock(autospec=True, return_value=True),
    )
    patch_slack_client_run_until = patch.object(
        slack_client, "_run_until", MagicMock(autospec=True, side_effect=[True, False])
    )
    patch_slack_client_run_command_async = patch.object(
        slack_client,
        "run_command_async",
        MagicMock(autospec=True, return_value="20221027001127600438"),
    )
    patch_slack_client_get_jobs_from_runner = patch.object(
        slack_client,
        "get_jobs_from_runner",
        MagicMock(autospec=True, return_value=mock_job_status),
    )

    upload_calls = call(
        channels="C02QY11UQ",
        content="minion:\n    True",
        filename="salt-results-20221027090136014442.yaml",
    )

    chat_postMessage_calls = [
        call(
            channel="C02QY11UQ",
            text="@garethgreenaway's job is submitted as salt jid 20221027001127600438",
        ),
        call(
            channel="C02QY11UQ",
            text="@garethgreenaway's job `['test.ping']` (id: 20221027001127600438) (target: {'target': 'minion', 'tgt_type': 'glob'}) returned",
        ),
    ]

    #
    # test with control as True and fire_all as False
    #
    with (
        patch_slack_client_run_until
    ), (
        patch_slack_client_run_command_async
    ), (
        patch_slack_client_get_jobs_from_runner
    ), patch_app_client_files_upload as app_client_files_upload, patch_app_client_chat_postMessage as app_client_chat_postMessage:
        slack_client.run_commands_from_slack_async(
            message_generator=message_generator,
            fire_all=False,
            tag="salt/engines/slack",
            control=True,
        )
        app_client_files_upload.asser_has_calls(upload_calls)
        app_client_chat_postMessage.asser_has_calls(chat_postMessage_calls)

    #
    # test with control and fire_all as True
    #
    patch_slack_client_run_until = patch.object(
        slack_client, "_run_until", MagicMock(autospec=True, side_effect=[True, False])
    )

    mock_event_send = MagicMock(return_value=True)
    patch_event_send = patch.dict(
        slack_bolt_engine.__salt__, {"event.send": mock_event_send}
    )

    event_send_calls = [
        call(
            "salt/engines/slack/message",
            {
                "message_data": {
                    "client_msg_id": "c1d0c13d-5e78-431e-9921-4786a7d27543",
                    "type": "message",
                    "text": '!test.ping target="minion"',
                    "user": "U02QY11UJ",
                    "ts": "1666829486.542159",
                    "blocks": [
                        {
                            "type": "rich_text",
                            "block_id": "2vdy",
                            "elements": [
                                {
                                    "type": "rich_text_section",
                                    "elements": [
                                        {
                                            "type": "text",
                                            "text": '!test.ping target="minion"',
                                        }
                                    ],
                                }
                            ],
                        }
                    ],
                    "team": "T02QY11UG",
                    "channel": "C02QY11UQ",
                    "event_ts": "1666829486.542159",
                    "channel_type": "channel",
                },
                "channel": "C02QY11UQ",
                "user": "U02QY11UJ",
                "user_name": "garethgreenaway",
                "cmdline": ["test.ping"],
                "target": {"target": "minion", "tgt_type": "glob"},
            },
        )
    ]
    with (
        patch_slack_client_run_until
    ), (
        patch_slack_client_run_command_async
    ), (
        patch_slack_client_get_jobs_from_runner
    ), (
        patch_event_send
    ), patch_app_client_files_upload as app_client_files_upload, patch_app_client_chat_postMessage as app_client_chat_postMessage:
        slack_client.run_commands_from_slack_async(
            message_generator=message_generator,
            fire_all=True,
            tag="salt/engines/slack",
            control=True,
        )
        app_client_files_upload.asser_has_calls(upload_calls)
        app_client_chat_postMessage.asser_has_calls(chat_postMessage_calls)
        mock_event_send.asser_has_calls(event_send_calls)


def test_run_command_async(slack_client):
    """
    Test slack engine: test_run_command_async
    """

    msg = {
        "message_data": {
            "client_msg_id": "6c71d7f9-a44d-402f-8f9f-d1bb5b650853",
            "type": "message",
            "text": '!test.ping target="minion"',
            "user": "U02QY11UJ",
            "ts": "1667427929.764169",
            "blocks": [
                {
                    "type": "rich_text",
                    "block_id": "AjL",
                    "elements": [
                        {
                            "type": "rich_text_section",
                            "elements": [
                                {"type": "text", "text": '!test.ping target="minion"'}
                            ],
                        }
                    ],
                }
            ],
            "team": "T02QY11UG",
            "channel": "C02QY11UQ",
            "event_ts": "1667427929.764169",
            "channel_type": "channel",
        },
        "channel": "C02QY11UQ",
        "user": "U02QY11UJ",
        "user_name": "garethgreenaway",
        "cmdline": ["test.ping"],
        "target": {"target": "minion", "tgt_type": "glob"},
    }

    local_client_mock = MagicMock(autospec=True, return_value=MockLocalClient())
    patch_local_client = patch("salt.client.LocalClient", local_client_mock)

    local_client_cmd_async_mock = MagicMock(
        autospec=True, return_value={"jid": "20221027001127600438"}
    )
    patch_local_client_cmd_async = patch.object(
        MockLocalClient, "cmd_async", local_client_cmd_async_mock
    )

    expected_calls = [call("minion", "test.ping", arg=[], kwarg={}, tgt_type="glob")]
    with patch_local_client, patch_local_client_cmd_async as local_client_cmd_async:
        ret = slack_client.run_command_async(msg)
        local_client_cmd_async.assert_has_calls(expected_calls)

    msg = {
        "message_data": {
            "client_msg_id": "35f4783f-8913-4687-8f04-21182bcacd5a",
            "type": "message",
            "text": "!test.arg arg1 arg2 arg3 key1=value1 key2=value2",
            "user": "U02QY11UJ",
            "ts": "1667429460.576889",
            "blocks": [
                {
                    "type": "rich_text",
                    "block_id": "EAzTy",
                    "elements": [
                        {
                            "type": "rich_text_section",
                            "elements": [
                                {
                                    "type": "text",
                                    "text": "!test.arg arg1 arg2 arg3 key1=value1 key2=value2",
                                }
                            ],
                        }
                    ],
                }
            ],
            "team": "T02QY11UG",
            "channel": "C02QY11UQ",
            "event_ts": "1667429460.576889",
            "channel_type": "channel",
        },
        "channel": "C02QY11UQ",
        "user": "U02QY11UJ",
        "user_name": "garethgreenaway",
        "cmdline": ["test.arg", "arg1", "arg2", "arg3", "key1=value1", "key2=value2"],
        "target": {"target": "*", "tgt_type": "glob"},
    }

    runner_client_mock = MagicMock(autospec=True, return_value=MockRunnerClient())
    patch_runner_client = patch("salt.runner.RunnerClient", runner_client_mock)

    runner_client_asynchronous_mock = MagicMock(
        autospec=True, return_value={"jid": "20221027001127600438"}
    )
    patch_runner_client_asynchronous = patch.object(
        MockRunnerClient, "asynchronous", runner_client_asynchronous_mock
    )

    expected_calls = [
        call(
            "test.arg",
            {
                "arg": ["arg1", "arg2", "arg3"],
                "kwarg": {"key1": "value1", "key2": "value2"},
            },
        )
    ]
    with (
        patch_runner_client
    ), patch_runner_client_asynchronous as runner_client_asynchronous:
        ret = slack_client.run_command_async(msg)
        runner_client_asynchronous.assert_has_calls(expected_calls)



import pytest

import salt.utils.platform

pytestmark = [
    pytest.mark.windows_whitelisted,
]


@pytest.fixture(scope="module")
def non_root_account():
    with pytest.helpers.create_account() as account:
        yield account


@pytest.fixture
def shell():
    if salt.utils.platform.is_windows():
        return "cmd"
    else:
        return "bash"


@pytest.mark.skip_if_not_root
def test_exec_code_all(salt_call_cli, non_root_account, shell):
    ret = salt_call_cli.run(
        "cmd.exec_code_all", shell, "echo good", runas=non_root_account.username
    )
    assert ret.returncode == 0


@pytest.mark.skip_on_windows(reason="No vt on Windows")
def test_long_stdout(salt_cli, salt_minion):
    echo_str = "salt" * 1000
    ret = salt_cli.run(
        "cmd.run", f"echo {echo_str}", use_vt=True, minion_tgt=salt_minion.id
    )
    assert ret.returncode == 0
    assert len(ret.data.strip()) == len(echo_str)



import pytest

import salt.modules.mac_xattr as xattr
import salt.utils.mac_utils
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {xattr: {}}


def test_list():
    """
    Test xattr.list
    """
    expected = {"spongebob": "squarepants", "squidward": "patrick"}
    with patch.object(
        xattr, "read", MagicMock(side_effect=["squarepants", "patrick"])
    ), patch(
        "salt.utils.mac_utils.execute_return_result",
        MagicMock(return_value="spongebob\nsquidward"),
    ):
        assert xattr.list_("path/to/file") == expected


def test_list_missing():
    """
    Test listing attributes of a missing file
    """
    with patch(
        "salt.utils.mac_utils.execute_return_result",
        MagicMock(side_effect=CommandExecutionError("No such file")),
    ):
        pytest.raises(CommandExecutionError, xattr.list_, "/path/to/file")


def test_read():
    """
    Test reading a specific attribute from a file
    """
    with patch(
        "salt.utils.mac_utils.execute_return_result",
        MagicMock(return_value="expected results"),
    ):
        assert xattr.read("/path/to/file", "com.attr") == "expected results"


def test_read_hex():
    """
    Test reading a specific attribute from a file
    """
    with patch.object(
        salt.utils.mac_utils,
        "execute_return_result",
        MagicMock(return_value="expected results"),
    ) as mock:
        assert (
            xattr.read("/path/to/file", "com.attr", **{"hex": True})
            == "expected results"
        )
        mock.assert_called_once_with(["xattr", "-p", "-x", "com.attr", "/path/to/file"])


def test_read_missing():
    """
    Test reading a specific attribute from a file
    """
    with patch(
        "salt.utils.mac_utils.execute_return_result",
        MagicMock(side_effect=CommandExecutionError("No such file")),
    ):
        pytest.raises(CommandExecutionError, xattr.read, "/path/to/file", "attribute")


def test_read_not_decodeable():
    """
    Test reading an attribute which returns non-UTF-8 bytes
    """
    with patch(
        "salt.utils.mac_utils.execute_return_result",
        MagicMock(
            side_effect=UnicodeDecodeError("UTF-8", b"\xd1expected results", 0, 1, "")
        ),
    ):
        assert xattr.read("/path/to/file", "com.attr") == "�expected results"


def test_write():
    """
    Test writing a specific attribute to a file
    """
    mock_cmd = MagicMock(return_value="squarepants")
    with patch.object(xattr, "read", mock_cmd), patch(
        "salt.utils.mac_utils.execute_return_success", MagicMock(return_value=True)
    ):
        assert xattr.write("/path/to/file", "spongebob", "squarepants")


def test_write_missing():
    """
    Test writing a specific attribute to a file
    """
    with patch(
        "salt.utils.mac_utils.execute_return_success",
        MagicMock(side_effect=CommandExecutionError("No such file")),
    ):
        pytest.raises(
            CommandExecutionError,
            xattr.write,
            "/path/to/file",
            "attribute",
            "value",
        )


def test_delete():
    """
    Test deleting a specific attribute from a file
    """
    mock_cmd = MagicMock(return_value={"spongebob": "squarepants"})
    with patch.object(xattr, "list_", mock_cmd), patch(
        "salt.utils.mac_utils.execute_return_success", MagicMock(return_value=True)
    ):
        assert xattr.delete("/path/to/file", "attribute")


def test_delete_missing():
    """
    Test deleting a specific attribute from a file
    """
    with patch(
        "salt.utils.mac_utils.execute_return_success",
        MagicMock(side_effect=CommandExecutionError("No such file")),
    ):
        pytest.raises(CommandExecutionError, xattr.delete, "/path/to/file", "attribute")


def test_clear():
    """
    Test clearing all attributes on a file
    """
    mock_cmd = MagicMock(return_value={})
    with patch.object(xattr, "list_", mock_cmd), patch(
        "salt.utils.mac_utils.execute_return_success", MagicMock(return_value=True)
    ):
        assert xattr.clear("/path/to/file")


def test_clear_missing():
    """
    Test clearing all attributes on a file
    """
    with patch(
        "salt.utils.mac_utils.execute_return_success",
        MagicMock(side_effect=CommandExecutionError("No such file")),
    ):
        pytest.raises(CommandExecutionError, xattr.clear, "/path/to/file")



import logging
import shutil
import sys

import pytest
from saltfactories.utils.functional import Loaders

log = logging.getLogger(__name__)


@pytest.fixture(scope="package")
def minion_id():
    return "func-tests-minion-opts"


@pytest.fixture(scope="module")
def state_tree(tmp_path_factory):
    state_tree_path = tmp_path_factory.mktemp("state-tree-base")
    try:
        yield state_tree_path
    finally:
        shutil.rmtree(str(state_tree_path), ignore_errors=True)


@pytest.fixture(scope="module")
def state_tree_prod(tmp_path_factory):
    state_tree_path = tmp_path_factory.mktemp("state-tree-prod")
    try:
        yield state_tree_path
    finally:
        shutil.rmtree(str(state_tree_path), ignore_errors=True)


@pytest.fixture(scope="module")
def minion_config_defaults():
    """
    Functional test modules can provide this fixture to tweak the default configuration dictionary
    passed to the minion factory
    """
    return {}


@pytest.fixture(scope="module")
def minion_config_overrides():
    """
    Functional test modules can provide this fixture to tweak the configuration
    overrides dictionary passed to the minion factory
    """
    return {}


@pytest.fixture(scope="module")
def minion_opts(
    salt_factories,
    minion_id,
    state_tree,
    state_tree_prod,
    minion_config_defaults,
    minion_config_overrides,
):
    minion_config_overrides.update(
        {
            "file_client": "local",
            "file_roots": {
                "base": [
                    str(state_tree),
                ],
                "prod": [
                    str(state_tree_prod),
                ],
            },
        }
    )

    if sys.platform.startswith("win"):
        # We need to set up winrepo on Windows
        minion_config_overrides.update(
            {
                "winrepo_source_dir": "salt://winrepo_ng",
                "winrepo_dir_ng": str(state_tree / "winrepo_ng"),
                "winrepo_dir": str(state_tree / "winrepo"),
            }
        )

    factory = salt_factories.salt_minion_daemon(
        minion_id,
        defaults=minion_config_defaults or None,
        overrides=minion_config_overrides,
    )
    return factory.config.copy()


@pytest.fixture(scope="module")
def master_config_defaults():
    """
    Functional test modules can provide this fixture to tweak the default configuration dictionary
    passed to the master factory
    """
    return {}


@pytest.fixture(scope="module")
def master_config_overrides():
    """
    Functional test modules can provide this fixture to tweak the configuration
    overrides dictionary passed to the master factory
    """
    return {}


@pytest.fixture(scope="module")
def master_opts(
    salt_factories,
    state_tree,
    state_tree_prod,
    master_config_defaults,
    master_config_overrides,
):
    master_config_overrides.update(
        {
            "file_client": "local",
            "file_roots": {
                "base": [
                    str(state_tree),
                ],
                "prod": [
                    str(state_tree_prod),
                ],
            },
        }
    )
    factory = salt_factories.salt_master_daemon(
        "func-tests-master-opts",
        defaults=master_config_defaults or None,
        overrides=master_config_overrides,
    )
    return factory.config.copy()


@pytest.fixture(scope="module")
def loaders(minion_opts):
    return Loaders(minion_opts, loaded_base_name=f"{__name__}.loaded")


@pytest.fixture(autouse=True)
def reset_loaders_state(loaders):
    try:
        # Run the tests
        yield
    finally:
        # Reset the loaders state
        loaders.reset_state()



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.status as status
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {status: {}}


def test_loadavg():
    """
    Test to return the current load average for the specified minion.
    """
    name = "mymonitor"

    ret = {"name": name, "changes": {}, "result": True, "data": {}, "comment": ""}

    mock = MagicMock(return_value=[])
    with patch.dict(status.__salt__, {"status.loadavg": mock}):
        comt = "Requested load average mymonitor not available "
        ret.update({"comment": comt, "result": False})
        assert status.loadavg(name) == ret

    mock = MagicMock(return_value={name: 3})
    with patch.dict(status.__salt__, {"status.loadavg": mock}):
        comt = "Min must be less than max"
        ret.update({"comment": comt, "result": False})
        assert status.loadavg(name, 1, 5) == ret

        comt = "Load avg is below minimum of 4 at 3.0"
        ret.update({"comment": comt, "data": 3})
        assert status.loadavg(name, 5, 4) == ret

        comt = "Load avg above maximum of 2 at 3.0"
        ret.update({"comment": comt, "data": 3})
        assert status.loadavg(name, 2, 1) == ret

        comt = "Load avg in acceptable range"
        ret.update({"comment": comt, "result": True})
        assert status.loadavg(name, 3, 1) == ret


def test_process():
    """
    Test to return whether the specified signature
    is found in the process tree.
    """
    name = "mymonitor"

    ret = {"name": name, "changes": {}, "result": True, "data": {}, "comment": ""}

    mock = MagicMock(side_effect=[{}, {name: 1}])
    with patch.dict(status.__salt__, {"status.pid": mock}):
        comt = 'Process signature "mymonitor" not found '
        ret.update({"comment": comt, "result": False})
        assert status.process(name) == ret

        comt = 'Process signature "mymonitor" was found '
        ret.update({"comment": comt, "result": True, "data": {name: 1}})
        assert status.process(name) == ret



"""
Tests for the Git state
"""

import functools
import inspect
import logging
import os
import shutil
import socket
import string
import tempfile
import urllib.parse

import pytest

import salt.utils.files
import salt.utils.path
from salt.utils.versions import Version
from tests.support.case import ModuleCase
from tests.support.helpers import TstSuiteLoggingHandler, with_tempdir
from tests.support.mixins import SaltReturnAssertsMixin
from tests.support.runtests import RUNTIME_VARS

TEST_REPO = "https://github.com/saltstack/salt-test-repo.git"


def __check_git_version(caller, min_version, skip_msg):
    """
    Common logic for version check
    """
    if inspect.isclass(caller):
        actual_setup = getattr(caller, "setUp", None)

        def setUp(self, *args, **kwargs):
            if not salt.utils.path.which("git"):
                self.skipTest("git is not installed")
            git_version = self.run_function("git.version")
            if Version(git_version) < Version(min_version):
                self.skipTest(skip_msg.format(min_version, git_version))
            if actual_setup is not None:
                actual_setup(self, *args, **kwargs)

        caller.setUp = setUp
        return caller

    @functools.wraps(caller)
    def wrapper(self, *args, **kwargs):
        if not salt.utils.path.which("git"):
            self.skipTest("git is not installed")
        git_version = self.run_function("git.version")
        if Version(git_version) < Version(min_version):
            self.skipTest(skip_msg.format(min_version, git_version))
        return caller(self, *args, **kwargs)

    return wrapper


def ensure_min_git(caller=None, min_version="1.6.5"):
    """
    Skip test if minimum supported git version is not installed
    """
    if caller is None:
        return functools.partial(ensure_min_git, min_version=min_version)

    return __check_git_version(
        caller, min_version, "git {0} or newer required to run this test (detected {1})"
    )


def uses_git_opts(caller):
    """
    Skip test if git_opts is not supported

    IMPORTANT! This decorator should be at the bottom of any decorators added
    to a given function.
    """
    min_version = "1.7.2"
    return __check_git_version(
        caller,
        min_version,
        "git_opts only supported in git {0} and newer (detected {1})",
    )


class WithGitMirror:
    def __init__(self, repo_url, **kwargs):
        self.repo_url = repo_url
        if "dir" not in kwargs:
            kwargs["dir"] = RUNTIME_VARS.TMP
        self.kwargs = kwargs

    def __call__(self, func):
        self.func = func
        return functools.wraps(func)(
            # pylint: disable=unnecessary-lambda
            lambda testcase, *args, **kwargs: self.wrap(testcase, *args, **kwargs)
            # pylint: enable=unnecessary-lambda
        )

    def wrap(self, testcase, *args, **kwargs):
        # Get temp dir paths
        mirror_dir = tempfile.mkdtemp(**self.kwargs)
        admin_dir = tempfile.mkdtemp(**self.kwargs)
        clone_dir = tempfile.mkdtemp(**self.kwargs)
        # Clean up the directories, we want git to actually create them
        os.rmdir(mirror_dir)
        os.rmdir(admin_dir)
        os.rmdir(clone_dir)
        # Create a URL to clone
        mirror_url = "file://" + mirror_dir
        # Mirror the repo
        testcase.run_function("git.clone", [mirror_dir], url=TEST_REPO, opts="--mirror")
        # Make sure the directory for the mirror now exists
        assert os.path.exists(mirror_dir)
        # Clone to the admin dir
        ret = testcase.run_state("git.latest", name=mirror_url, target=admin_dir)
        ret = ret[next(iter(ret))]
        assert os.path.exists(admin_dir)

        try:
            # Run the actual function with three arguments added:
            #   1. URL for the test to use to clone
            #   2. Cloned admin dir for making/pushing changes to the mirror
            #   3. Yet-nonexistent clone_dir for the test function to use as a
            #      destination for cloning.
            return self.func(
                testcase, mirror_url, admin_dir, clone_dir, *args, **kwargs
            )
        finally:
            shutil.rmtree(mirror_dir, ignore_errors=True)
            shutil.rmtree(admin_dir, ignore_errors=True)
            shutil.rmtree(clone_dir, ignore_errors=True)


with_git_mirror = WithGitMirror


@ensure_min_git
class GitTest(ModuleCase, SaltReturnAssertsMixin):
    """
    Validate the git state
    """

    def setUp(self):
        domain = urllib.parse.urlparse(TEST_REPO).netloc
        try:
            if hasattr(socket, "setdefaulttimeout"):
                # 10 second dns timeout
                socket.setdefaulttimeout(10)
            socket.gethostbyname(domain)
        except OSError:
            msg = "error resolving {0}, possible network issue?"
            self.skipTest(msg.format(domain))

    def tearDown(self):
        # Reset the dns timeout after the test is over
        socket.setdefaulttimeout(None)

    def _head(self, cwd):
        return self.run_function("git.rev_parse", [cwd, "HEAD"])

    @with_tempdir(create=False)
    @pytest.mark.slow_test
    def test_latest(self, target):
        """
        git.latest
        """
        ret = self.run_state("git.latest", name=TEST_REPO, target=target)
        self.assertSaltTrueReturn(ret)
        self.assertTrue(os.path.isdir(os.path.join(target, ".git")))

    @with_tempdir(create=False)
    @pytest.mark.slow_test
    def test_latest_config_get_regexp_retcode(self, target):
        """
        git.latest
        """

        handler = TstSuiteLoggingHandler(level=logging.DEBUG)
        ret_code_err = "failed with return code: 1"
        with handler:
            ret = self.run_state("git.latest", name=TEST_REPO, target=target)
            self.assertSaltTrueReturn(ret)
            self.assertTrue(os.path.isdir(os.path.join(target, ".git")))
            assert any(ret_code_err in s for s in handler.messages) is False

    @with_tempdir(create=False)
    @pytest.mark.slow_test
    def test_latest_with_rev_and_submodules(self, target):
        """
        git.latest
        """
        ret = self.run_state(
            "git.latest", name=TEST_REPO, rev="develop", target=target, submodules=True
        )
        self.assertSaltTrueReturn(ret)
        self.assertTrue(os.path.isdir(os.path.join(target, ".git")))

    @with_tempdir(create=False)
    @pytest.mark.slow_test
    def test_latest_failure(self, target):
        """
        git.latest
        """
        ret = self.run_state(
            "git.latest",
            name="https://youSpelledGitHubWrong.com/saltstack/salt-test-repo.git",
            rev="develop",
            target=target,
            submodules=True,
        )
        self.assertSaltFalseReturn(ret)
        self.assertFalse(os.path.isdir(os.path.join(target, ".git")))

    @with_tempdir()
    @pytest.mark.slow_test
    def test_latest_empty_dir(self, target):
        """
        git.latest
        """
        ret = self.run_state(
            "git.latest", name=TEST_REPO, rev="develop", target=target, submodules=True
        )
        self.assertSaltTrueReturn(ret)
        self.assertTrue(os.path.isdir(os.path.join(target, ".git")))

    @with_tempdir(create=False)
    @pytest.mark.slow_test
    def test_latest_unless_no_cwd_issue_6800(self, target):
        """
        cwd=target was being passed to _run_check which blew up if
        target dir did not already exist.
        """
        ret = self.run_state(
            "git.latest",
            name=TEST_REPO,
            rev="develop",
            target=target,
            unless=f"test -e {target}",
            submodules=True,
        )
        self.assertSaltTrueReturn(ret)
        self.assertTrue(os.path.isdir(os.path.join(target, ".git")))

    @with_tempdir(create=False)
    @pytest.mark.slow_test
    def test_numeric_rev(self, target):
        """
        git.latest with numeric revision
        """
        ret = self.run_state(
            "git.latest",
            name=TEST_REPO,
            rev=0.11,
            target=target,
            submodules=True,
            timeout=120,
        )
        self.assertSaltTrueReturn(ret)
        self.assertTrue(os.path.isdir(os.path.join(target, ".git")))

    @with_tempdir(create=False)
    @pytest.mark.slow_test
    def test_latest_with_local_changes(self, target):
        """
        Ensure that we fail the state when there are local changes and succeed
        when force_reset is True.
        """
        # Clone repo
        ret = self.run_state("git.latest", name=TEST_REPO, target=target)
        self.assertSaltTrueReturn(ret)
        self.assertTrue(os.path.isdir(os.path.join(target, ".git")))

        # Make change to LICENSE file.
        with salt.utils.files.fopen(os.path.join(target, "LICENSE"), "a") as fp_:
            fp_.write("Lorem ipsum dolor blah blah blah....\n")

        # Make sure that we now have uncommitted changes
        self.assertTrue(self.run_function("git.diff", [target, "HEAD"]))

        # Re-run state with force_reset=False
        ret = self.run_state(
            "git.latest", name=TEST_REPO, target=target, force_reset=False
        )
        self.assertSaltTrueReturn(ret)
        self.assertEqual(
            ret[next(iter(ret))]["comment"],
            "Repository {} is up-to-date, but with uncommitted changes. "
            "Set 'force_reset' to True to purge uncommitted changes.".format(target),
        )

        # Now run the state with force_reset=True
        ret = self.run_state(
            "git.latest", name=TEST_REPO, target=target, force_reset=True
        )
        self.assertSaltTrueReturn(ret)
        ret = ret[next(iter(ret))]
        # beginning of comment
        self.assertEqual(
            f"Repository {target} is up-to-date\n",
            ret["comment"][0 : 26 + len(target)],
        )
        self.assertIn("Uncommitted changes were discarded.", ret["comment"])
        # ending of comment
        self.assertEqual(
            "Repository was hard-reset to remote HEAD (bbefe81).", ret["comment"][-51:]
        )

        # Make sure that we no longer have uncommitted changes
        self.assertFalse(self.run_function("git.diff", [target, "HEAD"]))

    @with_git_mirror(TEST_REPO)
    @uses_git_opts
    @pytest.mark.slow_test
    def test_latest_fast_forward(self, mirror_url, admin_dir, clone_dir):
        """
        Test running git.latest state a second time after changes have been
        made to the remote repo.
        """
        # Clone the repo
        ret = self.run_state("git.latest", name=mirror_url, target=clone_dir)
        ret = ret[next(iter(ret))]
        assert ret["result"]

        # Make a change to the repo by editing the file in the admin copy
        # of the repo and committing.
        head_pre = self._head(admin_dir)
        with salt.utils.files.fopen(os.path.join(admin_dir, "LICENSE"), "a") as fp_:
            fp_.write("Hello world!")
        self.run_function(
            "git.commit",
            [admin_dir, "added a line"],
            git_opts='-c user.name="Foo Bar" -c user.email=foo@bar.com',
            opts="-a",
        )
        # Make sure HEAD is pointing to a new SHA so we know we properly
        # committed our change.
        head_post = self._head(admin_dir)
        assert head_pre != head_post

        # Push the change to the mirror
        # NOTE: the test will fail if the salt-test-repo's default branch
        # is changed.
        self.run_function("git.push", [admin_dir, "origin", "develop"])

        # Re-run the git.latest state on the clone_dir
        ret = self.run_state("git.latest", name=mirror_url, target=clone_dir)
        ret = ret[next(iter(ret))]
        assert ret["result"]

        # Make sure that the clone_dir now has the correct SHA
        assert head_post == self._head(clone_dir)

    @with_tempdir(create=False)
    def _changed_local_branch_helper(self, target, rev, hint):
        """
        We're testing two almost identical cases, the only thing that differs
        is the rev used for the git.latest state.
        """
        # Clone repo
        ret = self.run_state("git.latest", name=TEST_REPO, rev=rev, target=target)
        self.assertSaltTrueReturn(ret)

        # Check out a new branch in the clone and make a commit, to ensure
        # that when we re-run the state, it is not a fast-forward change
        self.run_function("git.checkout", [target, "new_branch"], opts="-b")
        with salt.utils.files.fopen(os.path.join(target, "foo"), "w"):
            pass
        self.run_function("git.add", [target, "."])
        self.run_function(
            "git.commit",
            [target, "add file"],
            git_opts='-c user.name="Foo Bar" -c user.email=foo@bar.com',
        )

        # Re-run the state, this should fail with a specific hint in the
        # comment field.
        ret = self.run_state("git.latest", name=TEST_REPO, rev=rev, target=target)
        self.assertSaltFalseReturn(ret)

        comment = ret[next(iter(ret))]["comment"]
        self.assertTrue(hint in comment)

    @uses_git_opts
    @pytest.mark.slow_test
    def test_latest_changed_local_branch_rev_head(self):
        """
        Test for presence of hint in failure message when the local branch has
        been changed and a the rev is set to HEAD

        This test will fail if the default branch for the salt-test-repo is
        ever changed.
        """
        self._changed_local_branch_helper(  # pylint: disable=no-value-for-parameter
            "HEAD",
            "The default remote branch (develop) differs from the local "
            "branch (new_branch)",
        )

    @uses_git_opts
    @pytest.mark.slow_test
    def test_latest_changed_local_branch_rev_develop(self):
        """
        Test for presence of hint in failure message when the local branch has
        been changed and a non-HEAD rev is specified
        """
        self._changed_local_branch_helper(  # pylint: disable=no-value-for-parameter
            "develop",
            "The desired rev (develop) differs from the name of the local "
            "branch (new_branch)",
        )

    @uses_git_opts
    @with_tempdir(create=False)
    @with_tempdir()
    @pytest.mark.slow_test
    def test_latest_updated_remote_rev(self, name, target):
        """
        Ensure that we don't exit early when checking for a fast-forward
        """
        # Initialize a new git repository
        self.run_function("git.init", [name])

        # Add and commit a file
        with salt.utils.files.fopen(os.path.join(name, "foo.txt"), "w") as fp_:
            fp_.write("Hello world\n")
        self.run_function("git.add", [name, "."])
        self.run_function(
            "git.commit",
            [name, "initial commit"],
            git_opts='-c user.name="Foo Bar" -c user.email=foo@bar.com',
        )

        # Run the state to clone the repo we just created
        ret = self.run_state(
            "git.latest",
            name=name,
            target=target,
        )
        self.assertSaltTrueReturn(ret)

        # Add another commit
        with salt.utils.files.fopen(os.path.join(name, "foo.txt"), "w") as fp_:
            fp_.write("Added a line\n")
        self.run_function(
            "git.commit",
            [name, "added a line"],
            git_opts='-c user.name="Foo Bar" -c user.email=foo@bar.com',
            opts="-a",
        )

        # Run the state again. It should pass, if it doesn't then there was
        # a problem checking whether or not the change is a fast-forward.
        ret = self.run_state(
            "git.latest",
            name=name,
            target=target,
        )
        self.assertSaltTrueReturn(ret)

    @with_tempdir(create=False)
    @pytest.mark.slow_test
    def test_latest_depth(self, target):
        """
        Test running git.latest state using the "depth" argument to limit the
        history. See #45394.
        """
        ret = self.run_state(
            "git.latest", name=TEST_REPO, rev="HEAD", target=target, depth=1
        )
        # HEAD is not a branch, this should fail
        self.assertSaltFalseReturn(ret)
        self.assertIn(
            "must be set to the name of a branch", ret[next(iter(ret))]["comment"]
        )

        ret = self.run_state(
            "git.latest",
            name=TEST_REPO,
            rev="non-default-branch",
            target=target,
            depth=1,
        )
        self.assertSaltTrueReturn(ret)
        self.assertTrue(os.path.isdir(os.path.join(target, ".git")))

    @with_git_mirror(TEST_REPO)
    @uses_git_opts
    @pytest.mark.slow_test
    def test_latest_sync_tags(self, mirror_url, admin_dir, clone_dir):
        """
        Test that a removed tag is properly reported as such and removed in the
        local clone, and that new tags are reported as new.
        """
        tag1 = "mytag1"
        tag2 = "mytag2"

        # Add and push a tag
        self.run_function("git.tag", [admin_dir, tag1])
        self.run_function("git.push", [admin_dir, "origin", tag1])

        # Clone the repo
        ret = self.run_state("git.latest", name=mirror_url, target=clone_dir)
        ret = ret[next(iter(ret))]
        assert ret["result"]

        # Now remove the tag
        self.run_function("git.push", [admin_dir, "origin", f":{tag1}"])
        # Add and push another tag
        self.run_function("git.tag", [admin_dir, tag2])
        self.run_function("git.push", [admin_dir, "origin", tag2])

        # Re-run the state with sync_tags=False. This should NOT delete the tag
        # from the local clone, but should report that a tag has been added.
        ret = self.run_state(
            "git.latest", name=mirror_url, target=clone_dir, sync_tags=False
        )
        ret = ret[next(iter(ret))]
        assert ret["result"]
        # Make ABSOLUTELY SURE both tags are present, since we shouldn't have
        # removed tag1.
        all_tags = self.run_function("git.list_tags", [clone_dir])
        assert tag1 in all_tags
        assert tag2 in all_tags
        # Make sure the reported changes are correct
        expected_changes = {"new_tags": [tag2]}
        assert ret["changes"] == expected_changes, ret["changes"]

        # Re-run the state with sync_tags=True. This should remove the local
        # tag, since it doesn't exist in the remote repository.
        ret = self.run_state(
            "git.latest", name=mirror_url, target=clone_dir, sync_tags=True
        )
        ret = ret[next(iter(ret))]
        assert ret["result"]
        # Make ABSOLUTELY SURE the expected tags are present/gone
        all_tags = self.run_function("git.list_tags", [clone_dir])
        assert tag1 not in all_tags
        assert tag2 in all_tags
        # Make sure the reported changes are correct
        expected_changes = {"deleted_tags": [tag1]}
        assert ret["changes"] == expected_changes, ret["changes"]

    @with_tempdir(create=False)
    @pytest.mark.slow_test
    def test_cloned(self, target):
        """
        Test git.cloned state
        """
        # Test mode
        ret = self.run_state("git.cloned", name=TEST_REPO, target=target, test=True)
        ret = ret[next(iter(ret))]
        assert ret["result"] is None
        assert ret["changes"] == {"new": f"{TEST_REPO} => {target}"}
        assert ret["comment"] == f"{TEST_REPO} would be cloned to {target}"

        # Now actually run the state
        ret = self.run_state("git.cloned", name=TEST_REPO, target=target)
        ret = ret[next(iter(ret))]
        assert ret["result"] is True
        assert ret["changes"] == {"new": f"{TEST_REPO} => {target}"}
        assert ret["comment"] == f"{TEST_REPO} cloned to {target}"

        # Run the state again to test idempotence
        ret = self.run_state("git.cloned", name=TEST_REPO, target=target)
        ret = ret[next(iter(ret))]
        assert ret["result"] is True
        assert not ret["changes"]
        assert ret["comment"] == f"Repository already exists at {target}"

        # Run the state again to test idempotence (test mode)
        ret = self.run_state("git.cloned", name=TEST_REPO, target=target, test=True)
        ret = ret[next(iter(ret))]
        assert not ret["changes"]
        assert ret["result"] is True
        assert ret["comment"] == f"Repository already exists at {target}"

    @with_tempdir(create=False)
    @pytest.mark.slow_test
    def test_cloned_with_branch(self, target):
        """
        Test git.cloned state with branch provided
        """
        old_branch = "master"
        new_branch = "develop"
        bad_branch = "thisbranchdoesnotexist"

        # Test mode
        ret = self.run_state(
            "git.cloned", name=TEST_REPO, target=target, branch=old_branch, test=True
        )
        ret = ret[next(iter(ret))]
        assert ret["result"] is None
        assert ret["changes"] == {"new": f"{TEST_REPO} => {target}"}
        assert ret["comment"] == "{} would be cloned to {} with branch '{}'".format(
            TEST_REPO, target, old_branch
        )

        # Now actually run the state
        ret = self.run_state(
            "git.cloned", name=TEST_REPO, target=target, branch=old_branch
        )
        ret = ret[next(iter(ret))]
        assert ret["result"] is True
        assert ret["changes"] == {"new": f"{TEST_REPO} => {target}"}
        assert ret["comment"] == "{} cloned to {} with branch '{}'".format(
            TEST_REPO, target, old_branch
        )

        # Run the state again to test idempotence
        ret = self.run_state(
            "git.cloned", name=TEST_REPO, target=target, branch=old_branch
        )
        ret = ret[next(iter(ret))]
        assert ret["result"] is True
        assert not ret["changes"]
        assert ret[
            "comment"
        ] == "Repository already exists at {} and is checked out to branch '{}'".format(
            target, old_branch
        )

        # Run the state again to test idempotence (test mode)
        ret = self.run_state(
            "git.cloned", name=TEST_REPO, target=target, test=True, branch=old_branch
        )
        ret = ret[next(iter(ret))]
        assert ret["result"] is True
        assert not ret["changes"]
        assert ret[
            "comment"
        ] == "Repository already exists at {} and is checked out to branch '{}'".format(
            target, old_branch
        )

        # Change branch (test mode)
        ret = self.run_state(
            "git.cloned", name=TEST_REPO, target=target, branch=new_branch, test=True
        )
        ret = ret[next(iter(ret))]
        assert ret["result"] is None
        assert ret["changes"] == {"branch": {"old": old_branch, "new": new_branch}}
        assert ret["comment"] == f"Branch would be changed to '{new_branch}'"

        # Now really change the branch
        ret = self.run_state(
            "git.cloned", name=TEST_REPO, target=target, branch=new_branch
        )
        ret = ret[next(iter(ret))]
        assert ret["result"] is True
        assert ret["changes"] == {"branch": {"old": old_branch, "new": new_branch}}
        assert ret["comment"] == f"Branch changed to '{new_branch}'"

        # Change back to original branch. This tests that we don't attempt to
        # checkout a new branch (i.e. git checkout -b) for a branch that exists
        # locally, as that would fail.
        ret = self.run_state(
            "git.cloned", name=TEST_REPO, target=target, branch=old_branch
        )
        ret = ret[next(iter(ret))]
        assert ret["result"] is True
        assert ret["changes"] == {"branch": {"old": new_branch, "new": old_branch}}
        assert ret["comment"] == f"Branch changed to '{old_branch}'"

        # Test switching to a nonexistent branch. This should fail.
        ret = self.run_state(
            "git.cloned", name=TEST_REPO, target=target, branch=bad_branch
        )
        ret = ret[next(iter(ret))]
        assert ret["result"] is False
        assert not ret["changes"]
        assert ret["comment"].startswith(f"Failed to change branch to '{bad_branch}':")

    @with_tempdir(create=False)
    @ensure_min_git(min_version="1.7.10")
    @pytest.mark.slow_test
    def test_cloned_with_nonexistant_branch(self, target):
        """
        Test git.cloned state with a nonexistent branch provided
        """
        branch = "thisbranchdoesnotexist"

        # Test mode
        ret = self.run_state(
            "git.cloned", name=TEST_REPO, target=target, branch=branch, test=True
        )
        ret = ret[next(iter(ret))]
        assert ret["result"] is None
        assert ret["changes"]
        assert ret["comment"] == "{} would be cloned to {} with branch '{}'".format(
            TEST_REPO, target, branch
        )

        # Now actually run the state
        ret = self.run_state("git.cloned", name=TEST_REPO, target=target, branch=branch)
        ret = ret[next(iter(ret))]
        assert ret["result"] is False
        assert not ret["changes"]
        assert ret["comment"].startswith("Clone failed:")
        assert "not found in upstream origin" in ret["comment"]

    @with_tempdir(create=False)
    @pytest.mark.slow_test
    def test_present(self, name):
        """
        git.present
        """
        ret = self.run_state("git.present", name=name, bare=True)
        self.assertSaltTrueReturn(ret)
        self.assertTrue(os.path.isfile(os.path.join(name, "HEAD")))

    @with_tempdir()
    @pytest.mark.slow_test
    def test_present_failure(self, name):
        """
        git.present
        """
        fname = os.path.join(name, "stoptheprocess")

        with salt.utils.files.fopen(fname, "a"):
            pass

        ret = self.run_state("git.present", name=name, bare=True)
        self.assertSaltFalseReturn(ret)
        self.assertFalse(os.path.isfile(os.path.join(name, "HEAD")))

    @with_tempdir()
    @pytest.mark.slow_test
    def test_present_empty_dir(self, name):
        """
        git.present
        """
        ret = self.run_state("git.present", name=name, bare=True)
        self.assertSaltTrueReturn(ret)
        self.assertTrue(os.path.isfile(os.path.join(name, "HEAD")))

    @with_tempdir()
    @pytest.mark.slow_test
    def test_config_set_value_with_space_character(self, name):
        """
        git.config
        """
        self.run_function("git.init", [name])

        ret = self.run_state(
            "git.config_set",
            name="user.name",
            value="foo bar",
            repo=name,
            **{"global": False},
        )
        self.assertSaltTrueReturn(ret)


@ensure_min_git
@uses_git_opts
class LocalRepoGitTest(ModuleCase, SaltReturnAssertsMixin):
    """
    Tests which do no require connectivity to github.com
    """

    def setUp(self):
        self.repo = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.admin = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        self.target = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        for dirname in (self.repo, self.admin, self.target):
            self.addCleanup(shutil.rmtree, dirname, ignore_errors=True)

        # Create bare repo
        self.run_function("git.init", [self.repo], bare=True)
        # Clone bare repo
        self.run_function("git.clone", [self.admin], url=self.repo)
        self._commit(self.admin, "", message="initial commit")
        self._push(self.admin)

    def _commit(self, repo_path, content, message):
        with salt.utils.files.fopen(os.path.join(repo_path, "foo"), "a") as fp_:
            fp_.write(content)
        self.run_function("git.add", [repo_path, "."])
        self.run_function(
            "git.commit",
            [repo_path, message],
            git_opts='-c user.name="Foo Bar" -c user.email=foo@bar.com',
        )

    def _push(self, repo_path, remote="origin", ref="master"):
        self.run_function("git.push", [repo_path], remote=remote, ref=ref)

    def _test_latest_force_reset_setup(self):
        # Perform the initial clone
        ret = self.run_state("git.latest", name=self.repo, target=self.target)
        self.assertSaltTrueReturn(ret)

        # Make and push changes to remote repo
        self._commit(self.admin, content="Hello world!\n", message="added a line")
        self._push(self.admin)

        # Make local changes to clone, but don't commit them
        with salt.utils.files.fopen(os.path.join(self.target, "foo"), "a") as fp_:
            fp_.write("Local changes!\n")

    @pytest.mark.slow_test
    def test_latest_force_reset_remote_changes(self):
        """
        This tests that an otherwise fast-forward change with local chanegs
        will not reset local changes when force_reset='remote_changes'
        """
        self._test_latest_force_reset_setup()

        # This should fail because of the local changes
        ret = self.run_state("git.latest", name=self.repo, target=self.target)
        self.assertSaltFalseReturn(ret)
        ret = ret[next(iter(ret))]
        self.assertIn("there are uncommitted changes", ret["comment"])
        self.assertIn("Set 'force_reset' to True (or 'remote-changes')", ret["comment"])
        self.assertEqual(ret["changes"], {})

        # Now run again with force_reset='remote_changes', the state should
        # succeed and discard the local changes
        ret = self.run_state(
            "git.latest",
            name=self.repo,
            target=self.target,
            force_reset="remote-changes",
        )
        self.assertSaltTrueReturn(ret)
        ret = ret[next(iter(ret))]
        self.assertIn("Uncommitted changes were discarded", ret["comment"])
        self.assertIn("Repository was fast-forwarded", ret["comment"])
        self.assertNotIn("forced update", ret["changes"])
        self.assertIn("revision", ret["changes"])

        # Add new local changes, but don't commit them
        with salt.utils.files.fopen(os.path.join(self.target, "foo"), "a") as fp_:
            fp_.write("More local changes!\n")

        # Now run again with force_reset='remote_changes', the state should
        # succeed with an up-to-date message and mention that there are local
        # changes, telling the user how to discard them.
        ret = self.run_state(
            "git.latest",
            name=self.repo,
            target=self.target,
            force_reset="remote-changes",
        )
        self.assertSaltTrueReturn(ret)
        ret = ret[next(iter(ret))]
        self.assertIn("up-to-date, but with uncommitted changes", ret["comment"])
        self.assertIn(
            "Set 'force_reset' to True to purge uncommitted changes", ret["comment"]
        )
        self.assertEqual(ret["changes"], {})

    @pytest.mark.slow_test
    def test_latest_force_reset_true_fast_forward(self):
        """
        This tests that an otherwise fast-forward change with local chanegs
        does reset local changes when force_reset=True
        """
        self._test_latest_force_reset_setup()

        # Test that local changes are discarded and that we fast-forward
        ret = self.run_state(
            "git.latest", name=self.repo, target=self.target, force_reset=True
        )
        self.assertSaltTrueReturn(ret)
        ret = ret[next(iter(ret))]
        self.assertIn("Uncommitted changes were discarded", ret["comment"])
        self.assertIn("Repository was fast-forwarded", ret["comment"])

        # Add new local changes
        with salt.utils.files.fopen(os.path.join(self.target, "foo"), "a") as fp_:
            fp_.write("More local changes!\n")

        # Running without setting force_reset should mention uncommitted changes
        ret = self.run_state("git.latest", name=self.repo, target=self.target)
        self.assertSaltTrueReturn(ret)
        ret = ret[next(iter(ret))]
        self.assertIn("up-to-date, but with uncommitted changes", ret["comment"])
        self.assertIn(
            "Set 'force_reset' to True to purge uncommitted changes", ret["comment"]
        )
        self.assertEqual(ret["changes"], {})

        # Test that local changes are discarded
        ret = self.run_state(
            "git.latest", name=TEST_REPO, target=self.target, force_reset=True
        )
        self.assertSaltTrueReturn(ret)
        ret = ret[next(iter(ret))]
        assert "Uncommitted changes were discarded" in ret["comment"]
        assert "Repository was hard-reset" in ret["comment"]
        assert "forced update" in ret["changes"]

    @pytest.mark.slow_test
    def test_latest_force_reset_true_non_fast_forward(self):
        """
        This tests that a non fast-forward change with divergent commits fails
        unless force_reset=True.
        """
        self._test_latest_force_reset_setup()

        # Reset to remote HEAD
        ret = self.run_state(
            "git.latest", name=self.repo, target=self.target, force_reset=True
        )
        self.assertSaltTrueReturn(ret)
        ret = ret[next(iter(ret))]
        self.assertIn("Uncommitted changes were discarded", ret["comment"])
        self.assertIn("Repository was fast-forwarded", ret["comment"])

        # Make and push changes to remote repo
        self._commit(self.admin, content="New line\n", message="added another line")
        self._push(self.admin)

        # Make different changes to local file and commit locally
        self._commit(
            self.target,
            content="Different new line\n",
            message="added a different line",
        )

        # This should fail since the local clone has diverged and cannot
        # fast-forward to the remote rev
        ret = self.run_state("git.latest", name=self.repo, target=self.target)
        self.assertSaltFalseReturn(ret)
        ret = ret[next(iter(ret))]
        self.assertIn("this is not a fast-forward merge", ret["comment"])
        self.assertIn("Set 'force_reset' to True to force this update", ret["comment"])
        self.assertEqual(ret["changes"], {})

        # Repeat the state with force_reset=True and confirm that the hard
        # reset was performed
        ret = self.run_state(
            "git.latest", name=self.repo, target=self.target, force_reset=True
        )
        self.assertSaltTrueReturn(ret)
        ret = ret[next(iter(ret))]
        self.assertIn("Repository was hard-reset", ret["comment"])
        self.assertIn("forced update", ret["changes"])
        self.assertIn("revision", ret["changes"])

    @pytest.mark.slow_test
    def test_renamed_default_branch(self):
        """
        Test the case where the remote branch has been removed
        https://github.com/saltstack/salt/issues/36242
        """
        # Rename remote 'master' branch to 'develop'
        os.rename(
            os.path.join(self.repo, "refs", "heads", "master"),
            os.path.join(self.repo, "refs", "heads", "develop"),
        )

        # Run git.latest state. This should successfully clone and fail with a
        # specific error in the comment field.
        ret = self.run_state(
            "git.latest",
            name=self.repo,
            target=self.target,
            rev="develop",
        )
        self.assertSaltFalseReturn(ret)
        self.assertEqual(
            ret[next(iter(ret))]["comment"],
            "Remote HEAD refers to a ref that does not exist. "
            "This can happen when the default branch on the "
            "remote repository is renamed or deleted. If you "
            "are unable to fix the remote repository, you can "
            "work around this by setting the 'branch' argument "
            "(which will ensure that the named branch is created "
            "if it does not already exist).\n\n"
            "Changes already made: {} cloned to {}".format(self.repo, self.target),
        )
        self.assertEqual(
            ret[next(iter(ret))]["changes"],
            {"new": f"{self.repo} => {self.target}"},
        )

        # Run git.latest state again. This should fail again, with a different
        # error in the comment field, and should not change anything.
        ret = self.run_state(
            "git.latest",
            name=self.repo,
            target=self.target,
            rev="develop",
        )
        self.assertSaltFalseReturn(ret)
        self.assertEqual(
            ret[next(iter(ret))]["comment"],
            "Cannot set/unset upstream tracking branch, local "
            "HEAD refers to nonexistent branch. This may have "
            "been caused by cloning a remote repository for which "
            "the default branch was renamed or deleted. If you "
            "are unable to fix the remote repository, you can "
            "work around this by setting the 'branch' argument "
            "(which will ensure that the named branch is created "
            "if it does not already exist).",
        )
        self.assertEqual(ret[next(iter(ret))]["changes"], {})

        # Run git.latest state again with a branch manually set. This should
        # checkout a new branch and the state should pass.
        ret = self.run_state(
            "git.latest",
            name=self.repo,
            target=self.target,
            rev="develop",
            branch="develop",
        )
        # State should succeed
        self.assertSaltTrueReturn(ret)
        self.assertSaltCommentRegexpMatches(
            ret,
            "New branch 'develop' was checked out, with origin/develop "
            r"\([0-9a-f]{7}\) as a starting point",
        )
        # Only the revision should be in the changes dict.
        self.assertEqual(list(ret[next(iter(ret))]["changes"].keys()), ["revision"])
        # Since the remote repo was incorrectly set up, the local head should
        # not exist (therefore the old revision should be None).
        self.assertEqual(ret[next(iter(ret))]["changes"]["revision"]["old"], None)
        # Make sure the new revision is a SHA (40 chars, all hex)
        self.assertTrue(len(ret[next(iter(ret))]["changes"]["revision"]["new"]) == 40)
        self.assertTrue(
            all(
                [
                    x in string.hexdigits
                    for x in ret[next(iter(ret))]["changes"]["revision"]["new"]
                ]
            )
        )



"""
    :codeauthor: Justin Anderson <janderson@saltstack.com>
"""

import pathlib
import shutil

import attr
import pytest

from tests.support.helpers import PRE_PYTEST_SKIP_OR_NOT

pytestmark = [
    pytest.mark.core_test,
    pytest.mark.windows_whitelisted,
    pytest.mark.timeout_unless_on_windows(200),
]


@attr.s(frozen=True, slots=True)
class SaltCallCliWrapper:
    salt_call_cli = attr.ib()

    def run(self, *args, **kwargs):
        beacon_event_timeout = 120
        cli_run_timeout = beacon_event_timeout + 30
        return self.salt_call_cli.run(
            *args, timeout=beacon_event_timeout, _timeout=cli_run_timeout, **kwargs
        )


@pytest.fixture(scope="module")
def salt_call_cli(salt_call_cli):
    return SaltCallCliWrapper(salt_call_cli)


@pytest.fixture(scope="module")
def cleanup_beacons_config_module(salt_minion, salt_call_cli):
    minion_conf_d_dir = (
        pathlib.Path(salt_minion.config_dir)
        / pathlib.Path(salt_minion.config["default_include"]).parent
    )
    if not minion_conf_d_dir.is_dir():
        minion_conf_d_dir.mkdir()
    beacons_config_file_path = minion_conf_d_dir / "beacons.conf"
    ret = salt_call_cli.run("beacons.reset")
    assert ret.returncode == 0
    assert ret.data
    assert ret.data["result"] is True
    if beacons_config_file_path.exists():
        beacons_config_file_path.unlink()
    try:
        yield beacons_config_file_path
    finally:
        if beacons_config_file_path.exists():
            beacons_config_file_path.unlink()


@pytest.fixture(autouse=True)
def cleanup_beacons_config(cleanup_beacons_config_module, salt_call_cli):
    try:
        yield cleanup_beacons_config_module
    finally:
        ret = salt_call_cli.run("beacons.reset")
        assert ret.returncode == 0
        assert ret.data
        assert ret.data["result"] is True


@pytest.fixture(scope="module")
def inotify_file_path(tmp_path_factory):
    inotify_directory = tmp_path_factory.mktemp("important-files")
    try:
        yield inotify_directory / "really-important"
    finally:
        shutil.rmtree(str(inotify_directory), ignore_errors=True)


@pytest.fixture(scope="module")
def pillar_tree(
    base_env_pillar_tree_root_dir, salt_minion, salt_call_cli, inotify_file_path
):
    top_file = """
    base:
      '{}':
        - beacons
    """.format(
        salt_minion.id
    )
    beacon_pillar_file = """
    beacons:
      inotify:
        - files:
            {}:
              mask:
                - open
                - create
                - close_write
    """.format(
        inotify_file_path
    )
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_pillar_tree_root_dir
    )
    beacon_tempfile = pytest.helpers.temp_file(
        "beacons.sls", beacon_pillar_file, base_env_pillar_tree_root_dir
    )
    try:
        with top_tempfile, beacon_tempfile:
            ret = salt_call_cli.run("saltutil.refresh_pillar", wait=True)
            assert ret.returncode == 0
            assert ret.data is True
            yield
    finally:
        # Refresh pillar again to cleaup the temp pillar
        ret = salt_call_cli.run("saltutil.refresh_pillar", wait=True)
        assert ret.returncode == 0
        assert ret.data is True


@attr.s(frozen=True, slots=True)
class Beacon:
    name = attr.ib()
    data = attr.ib()


def beacon_instance_ids(value):
    return str(value)


@pytest.fixture(
    params=[
        Beacon("ps", [{"processes": {"apache2": "stopped"}}]),
        Beacon(
            "watch_apache",
            [{"processes": {"apache2": "stopped"}}, {"beacon_module": "ps"}],
        ),
    ],
    ids=beacon_instance_ids,
)
def beacon_instance(request):
    return request.param


def test_add_and_delete(salt_call_cli, beacon_instance):
    """
    Test adding and deleting a beacon
    """
    # Add the beacon
    ret = salt_call_cli.run(
        "beacons.add", beacon_instance.name, beacon_data=beacon_instance.data
    )
    assert ret.returncode == 0
    assert ret.data
    assert ret.data["result"] is True

    # Save beacons
    ret = salt_call_cli.run("beacons.save")
    assert ret.returncode == 0
    assert ret.data
    assert ret.data["result"] is True

    # Delete beacon
    ret = salt_call_cli.run("beacons.delete", beacon_instance.name)
    assert ret.returncode == 0
    assert ret.data
    assert ret.data["result"] is True


@pytest.fixture
def beacon(beacon_instance, salt_call_cli):
    ret = salt_call_cli.run(
        "beacons.add", beacon_instance.name, beacon_data=beacon_instance.data
    )
    assert ret.returncode == 0
    assert ret.data
    assert ret.data["result"] is True

    # Save beacons
    ret = salt_call_cli.run("beacons.save")
    assert ret.returncode == 0
    assert ret.data
    assert ret.data["result"] is True

    # assert beacon exists
    ret = salt_call_cli.run("beacons.list", return_yaml=False)
    assert ret.returncode == 0
    assert ret.data
    assert beacon_instance.name in ret.data

    yield beacon_instance


def test_disable(salt_call_cli, beacon):
    """
    Test disabling beacons
    """
    ret = salt_call_cli.run("beacons.disable")
    assert ret.returncode == 0
    assert ret.data
    assert ret.data["result"] is True

    # assert beacons are disabled
    ret = salt_call_cli.run("beacons.list", return_yaml=False)
    assert ret.returncode == 0
    assert ret.data
    assert ret.data["enabled"] is False

    # disable added beacon
    ret = salt_call_cli.run("beacons.disable_beacon", beacon.name)
    assert ret.returncode == 0
    assert ret.data
    assert ret.data["result"] is True

    # assert beacon is disabled
    ret = salt_call_cli.run("beacons.list", return_yaml=False)
    assert ret.returncode == 0
    assert ret.data
    assert beacon.name in ret.data
    for beacon_data in ret.data[beacon.name]:
        if "enabled" in beacon_data:
            assert beacon_data["enabled"] is False
            break
    else:
        pytest.fail("Did not find the beacon data with the 'enabled' key")


@pytest.fixture
def disabled_beacon(beacon, salt_call_cli):
    ret = salt_call_cli.run("beacons.disable")
    assert ret.returncode == 0
    assert ret.data
    assert ret.data["result"] is True
    return beacon


def test_enable(salt_call_cli, disabled_beacon):
    """
    Test enabling beacons
    """
    # enable beacons on minion
    ret = salt_call_cli.run("beacons.enable")
    assert ret.returncode == 0
    assert ret.data
    assert ret.data["result"] is True

    # assert beacons are enabled
    ret = salt_call_cli.run("beacons.list", return_yaml=False)
    assert ret.returncode == 0
    assert ret.data
    assert ret.data["enabled"] is True


@pytest.mark.skipif(
    PRE_PYTEST_SKIP_OR_NOT,
    reason=(
        "Skip until https://github.com/saltstack/salt/issues/31516 problems are resolved."
    ),
)
def test_enabled_beacons(salt_call_cli, beacon):
    """
    Test enabled specific beacon
    """
    # enable added beacon
    ret = salt_call_cli.run("beacons.enable_beacon", beacon.name)
    assert ret.returncode == 0
    assert ret.data
    assert ret.data["result"] is True

    # assert beacon ps is enabled
    ret = salt_call_cli.run("beacons.list", return_yaml=False)
    assert ret.returncode == 0
    assert ret.data
    assert ret.data["enabled"] is True
    assert beacon.name in ret.data
    for beacon_data in ret.data[beacon.name]:
        if "enabled" in beacon_data:
            assert beacon_data["enabled"] is False
            break
    else:
        pytest.fail("Did not find the beacon data with the 'enabled' key")


@pytest.mark.usefixtures("pillar_tree")
def test_list(salt_call_cli, beacon, inotify_file_path):
    """
    Test listing the beacons
    """
    # list beacons
    ret = salt_call_cli.run("beacons.list", return_yaml=False)
    assert ret.returncode == 0
    assert ret.data
    assert ret.data == {
        beacon.name: beacon.data,
        "inotify": [
            {
                "files": {
                    str(inotify_file_path): {"mask": ["open", "create", "close_write"]}
                }
            }
        ],
    }


@pytest.mark.usefixtures("pillar_tree")
def test_list_only_include_opts(salt_call_cli, beacon):
    """
    Test listing the beacons which only exist in opts

    When beacon.save is used to save the running beacons to
    a file, it uses beacons.list to get that list and should
    only return those from opts and not pillar.

    In this test, we're making sure we get only get back the
    beacons that are in opts and not those in pillar.
    """
    # list beacons
    ret = salt_call_cli.run(
        "beacons.list", return_yaml=False, include_opts=True, include_pillar=False
    )
    assert ret.returncode == 0
    assert ret.data
    assert ret.data == {beacon.name: beacon.data}


@pytest.mark.usefixtures("pillar_tree", "beacon")
def test_list_only_include_pillar(salt_call_cli, inotify_file_path):
    """
    Test listing the beacons which only exist in pillar
    """
    # list beacons
    ret = salt_call_cli.run(
        "beacons.list", return_yaml=False, include_opts=False, include_pillar=True
    )
    assert ret.returncode == 0
    assert ret.data
    assert ret.data == {
        "inotify": [
            {
                "files": {
                    str(inotify_file_path): {"mask": ["open", "create", "close_write"]}
                }
            }
        ]
    }


def test_list_available(salt_call_cli):
    """
    Test listing the beacons
    """
    # list beacons
    ret = salt_call_cli.run("beacons.list_available", return_yaml=False)
    assert ret.returncode == 0
    assert ret.data
    assert "ps" in ret.data
